#include "Builder.hpp"
#include <random>

using std::string;
using std::vector;

class BuildCell
{
public:
    uint32_t m_visited = 0;
    Point m_point;
};

bool build(const Puzzle& puzzle,
           Board* pBoard,
           Point* pStartPoint,
           string* pMsg)
{
    std::random_device rd;
    std::default_random_engine gen(rd());

    uint32_t width = puzzle.m_tSize;
    uint32_t height = puzzle.m_tSize;

    uint32_t minArea = (width * height + 4) / 5;
    uint32_t area = 0;

    Board board(width, height);
    for (uint32_t y = 0; y < height; y++)
    {
        for (uint32_t x = 0; x < width; x++)
        {
            board(x, y) = '?';
        }
    }

    for (auto point : puzzle.m_oSqs)
    {
        board[point] = '#';
    }

    vector<Point> inPoints(puzzle.m_iSqs);

    {
        std::uniform_int_distribution<uint32_t> xdist(0, width - 1);
        std::uniform_int_distribution<uint32_t> ydist(0, height - 1);
        uint32_t successes = 0;
        for (uint32_t attempt = 0; attempt < 10; attempt++)
        {
            uint32_t x = xdist(gen);
            uint32_t y = 0;
            if (board(x, y) == ' ')
            {
                successes++;
                break;
            }
            else if (board(x, y) != '#')
            {
                inPoints.emplace_back(x, y);
                successes++;
                break;
            }
        }
        for (uint32_t attempt = 0; attempt < 10; attempt++)
        {
            uint32_t x = xdist(gen);
            uint32_t y = height - 1;
            if (board(x, y) == ' ')
            {
                successes++;
                break;
            }
            else if (board(x, y) != '#')
            {
                inPoints.emplace_back(x, y);
                successes++;
                break;
            }
        }
        for (uint32_t attempt = 0; attempt < 10; attempt++)
        {
            uint32_t x = 0;
            uint32_t y = ydist(gen);
            if (board(x, y) == ' ')
            {
                successes++;
                break;
            }
            else if (board(x, y) != '#')
            {
                inPoints.emplace_back(x, y);
                successes++;
                break;
            }
        }
        for (uint32_t attempt = 0; attempt < 10; attempt++)
        {
            uint32_t x = width - 1;
            uint32_t y = ydist(gen);
            if (board(x, y) == ' ')
            {
                successes++;
                break;
            }
            else if (board(x, y) != '#')
            {
                inPoints.emplace_back(x, y);
                successes++;
                break;
            }
        }
        if (successes != 4)
        {
            if (pMsg) *pMsg = "Failed to extend to borders";
            return false;
        }
    }

    std::shuffle(inPoints.begin(), inPoints.end(), gen);

    for (auto point : inPoints)
    {
        if (board[point] == '#')
        {
            if (pMsg) *pMsg = "Point is both in and out";
            return false;
        }
    }

    vector<Point> allIn;

    size_t inPos = 0;
    if (inPos < inPoints.size())
    {
        board[inPoints[inPos]] = ' ';
        allIn.push_back(inPoints[inPos]);
        inPos++;
        area++;
    }

    Grid<BuildCell> grid(width, height);
    uint32_t v = 0;
    vector<Point> oldPoints;
    vector<Point> newPoints;

    while (inPos < inPoints.size())
    {
        Point inPoint = inPoints[inPos];
        inPos++;

        if (board[inPoint] == ' ')
        {
            continue;
        }

        v++;

        oldPoints.push_back(inPoint);
        grid[inPoint].m_visited = v;
        grid[inPoint].m_point = inPoint;

        bool foundPath = false;
        Point foundPoint;

        while (!oldPoints.empty())
        {
            for (auto oldPoint : oldPoints)
            {
                for (uint32_t dir = 0; dir < 4; dir++)
                {
                    Delta delta;
                    if (dir == 0) { delta = Delta(0, 1); }
                    else if (dir == 1) { delta = Delta(0, -1); }
                    else if (dir == 2) { delta = Delta(-1, 0); }
                    else if (dir == 3) { delta = Delta(1, 0); }

                    Point newPoint = oldPoint + delta;
                    if (newPoint.m_x >= width ||
                        newPoint.m_y >= height ||
                        board[newPoint] == '#' ||
                        grid[newPoint].m_visited == v)
                    {
                        continue;
                    }

                    newPoints.push_back(newPoint);
                    grid[newPoint].m_visited = v;
                    grid[newPoint].m_point = oldPoint;

                    if (board[newPoint] == ' ')
                    {
                        foundPath = true;
                        foundPoint = newPoint;
                        newPoints.clear();
                        break;
                    }
                }

                if (foundPath)
                {
                    break;
                }
            }

            oldPoints.clear();
            oldPoints.swap(newPoints);

            if (foundPath)
            {
                break;
            }
        }

        if (!foundPath)
        {
            if (pMsg) *pMsg = "Path not found";
            return false;
        }

        Point tmpPoint = foundPoint;
        while (true)
        {
            auto& cell = grid[tmpPoint];
            if (cell.m_point == tmpPoint)
            {
                break;
            }
            tmpPoint = cell.m_point;
            board[tmpPoint] = ' ';
            allIn.push_back(tmpPoint);
            area++;
        }
    }

    while (area < minArea)
    {
        std::uniform_int_distribution<uint32_t> dist(0, (uint32_t)allIn.size() - 1);
        uint32_t iIn = dist(gen);
        Point inPoint = allIn[iIn];

        uint32_t maxSqSize = 0;
        uint32_t sqSize = 0;
        while (true)
        {
            sqSize++;
            uint32_t lowX = inPoint.m_x - sqSize;
            uint32_t highX = inPoint.m_x + sqSize;
            uint32_t lowY = inPoint.m_y - sqSize;
            uint32_t highY = inPoint.m_y + sqSize;
            if (lowX >= width ||
                highX >= width ||
                lowY >= height ||
                highY >= height)
            {
                break;
            }
            bool ok = true;
            uint32_t inCount = 0;
            for (uint32_t y = lowY; y <= highY; y++)
            {
                for (uint32_t x = lowX; x <= highX; x++)
                {
                    if (board(x, y) == '#')
                    {
                        ok = false;
                        break;
                    }
                    else if (board(x, y) == ' ')
                    {
                        inCount++;
                    }
                }
                if (!ok) break;
            }
            if (!ok) break;

            v++;
            oldPoints.push_back(inPoint);
            grid[inPoint].m_visited = v;
            grid[inPoint].m_point = inPoint;
            uint32_t foundCount = 1;

            while (!oldPoints.empty())
            {
                for (auto oldPoint : oldPoints)
                {
                    for (uint32_t dir = 0; dir < 4; dir++)
                    {
                        Delta delta;
                        if (dir == 0) { delta = Delta(0, 1); }
                        else if (dir == 1) { delta = Delta(0, -1); }
                        else if (dir == 2) { delta = Delta(-1, 0); }
                        else if (dir == 3) { delta = Delta(1, 0); }

                        Point newPoint = oldPoint + delta;
                        if (newPoint.m_x < lowX ||
                            newPoint.m_x > highX ||
                            newPoint.m_y < lowY ||
                            newPoint.m_y > highY ||
                            board[newPoint] != ' ' ||
                            grid[newPoint].m_visited == v)
                        {
                            continue;
                        }

                        newPoints.push_back(newPoint);
                        grid[newPoint].m_visited = v;
                        grid[newPoint].m_point = oldPoint;
                        foundCount++;
                    }
                }

                oldPoints.clear();
                oldPoints.swap(newPoints);
            }

            if (foundCount != inCount)
            {
                break;
            }

            maxSqSize = sqSize - 1;
        }

        sqSize = maxSqSize;
        uint32_t lowX = inPoint.m_x - sqSize;
        uint32_t highX = inPoint.m_x + sqSize;
        uint32_t lowY = inPoint.m_y - sqSize;
        uint32_t highY = inPoint.m_y + sqSize;
        for (uint32_t y = lowY; y <= highY; y++)
        {
            for (uint32_t x = lowX; x <= highX; x++)
            {
                if (board(x, y) != ' ')
                {
                    board(x, y) = ' ';
                    allIn.emplace_back(x, y);
                    area++;
                }
            }
        }
    }

    uint32_t checkArea = 0;
    for (uint32_t y = 0; y < height; y++)
    {
        for (uint32_t x = 0; x < width; x++)
        {
            if (board(x, y) == ' ')
            {
                checkArea++;
            }
        }
    }

    if (checkArea != area)
    {
        if (pMsg) *pMsg = "area mismatch";
        return false;
    }

    while (true)
    {
        uint32_t vertexCount = 0;
        for (uint32_t y = 0; y <= height; y++)
        {
            for (uint32_t x = 0; x <= width; x++)
            {
                uint32_t cornerSW =
                    (x - 1 < width &&
                     y - 1 < height &&
                     board(x - 1, y - 1) == ' ');
                uint32_t cornerSE =
                             (x < width &&
                              y - 1 < height &&
                              board(x, y - 1) == ' ');
                uint32_t cornerNW =
                                  (x - 1 < width &&
                                   y < height &&
                                   board(x - 1, y) == ' ');
                uint32_t cornerNE =
                                           (x < width &&
                                            y < height &&
                                            board(x, y) == ' ');
                uint32_t cornerCount = cornerSW + cornerSE + cornerNW + cornerNE;
                if (cornerCount == 1 || cornerCount == 3)
                {
                    vertexCount++;
                }
                if (cornerCount == 2 && cornerSW == cornerNE)
                {
                    if (pMsg) *pMsg = "Corners touching";
                    return false;
                }
            }
        }

        if (vertexCount > puzzle.m_vMax)
        {
            if (pMsg) *pMsg = "Too many vertices";
            return false;
        }

        if (vertexCount >= puzzle.m_vMin)
        {
            break;
        }

        std::uniform_int_distribution<uint32_t> xdist(0, width - 1);
        std::uniform_int_distribution<uint32_t> ydist(0, height - 1);

        Point inPoint(xdist(gen), ydist(gen));

        if (board[inPoint] != '?')
        {
            continue;
        }

        v++;

        oldPoints.push_back(inPoint);
        grid[inPoint].m_visited = v;
        grid[inPoint].m_point = inPoint;

        bool foundPath = false;
        Point foundPoint;

        while (!oldPoints.empty())
        {
            for (auto oldPoint : oldPoints)
            {
                for (uint32_t dir = 0; dir < 4; dir++)
                {
                    Delta delta;
                    if (dir == 0) { delta = Delta(0, 1); }
                    else if (dir == 1) { delta = Delta(0, -1); }
                    else if (dir == 2) { delta = Delta(-1, 0); }
                    else if (dir == 3) { delta = Delta(1, 0); }

                    Point newPoint = oldPoint + delta;
                    if (newPoint.m_x >= width ||
                        newPoint.m_y >= height ||
                        board[newPoint] == '#' ||
                        grid[newPoint].m_visited == v)
                    {
                        continue;
                    }

                    newPoints.push_back(newPoint);
                    grid[newPoint].m_visited = v;
                    grid[newPoint].m_point = oldPoint;

                    if (board[newPoint] == ' ')
                    {
                        foundPath = true;
                        foundPoint = newPoint;
                        newPoints.clear();
                        break;
                    }
                }

                if (foundPath)
                {
                    break;
                }
            }

            oldPoints.clear();
            oldPoints.swap(newPoints);

            if (foundPath)
            {
                break;
            }
        }

        if (!foundPath)
        {
            if (pMsg) *pMsg = "Path not found (vertices)";
            return false;
        }

        Point tmpPoint = foundPoint;
        while (true)
        {
            auto& cell = grid[tmpPoint];
            if (cell.m_point == tmpPoint)
            {
                break;
            }
            tmpPoint = cell.m_point;
            board[tmpPoint] = ' ';
            allIn.push_back(tmpPoint);
            area++;
        }
    }

    for (uint32_t y = 0; y < height; y++)
    {
        for (uint32_t x = 0; x < width; x++)
        {
            if (board(x, y) == '?')
            {
                board(x, y) = '#';
            }
        }
    }

    {
        uint32_t outCount = 0;
        for (uint32_t y = 0; y < height; y++)
        {
            for (uint32_t x = 0; x < width; x++)
            {
                if (board(x, y) == '#')
                {
                    outCount++;
                }
            }
        }

        uint32_t foundCount = 0;

        v++;
        for (uint32_t y = 0; y < height; y++)
        {
            {
                Point outPoint(0, y);
                if (board[outPoint] == '#')
                {
                    oldPoints.push_back(outPoint);
                    grid[outPoint].m_visited = v;
                    grid[outPoint].m_point = outPoint;
                    foundCount++;
                }
            }
            {
                Point outPoint(width - 1, y);
                if (board[outPoint] == '#')
                {
                    oldPoints.push_back(outPoint);
                    grid[outPoint].m_visited = v;
                    grid[outPoint].m_point = outPoint;
                    foundCount++;
                }
            }
        }
        for (uint32_t x = 1; x + 1 < width; x++)
        {
            {
                Point outPoint(x, 0);
                if (board[outPoint] == '#')
                {
                    oldPoints.push_back(outPoint);
                    grid[outPoint].m_visited = v;
                    grid[outPoint].m_point = outPoint;
                    foundCount++;
                }
            }
            {
                Point outPoint(x, height - 1);
                if (board[outPoint] == '#')
                {
                    oldPoints.push_back(outPoint);
                    grid[outPoint].m_visited = v;
                    grid[outPoint].m_point = outPoint;
                    foundCount++;
                }
            }
        }

        while (!oldPoints.empty())
        {
            for (auto oldPoint : oldPoints)
            {
                for (uint32_t dir = 0; dir < 4; dir++)
                {
                    Delta delta;
                    if (dir == 0) { delta = Delta(0, 1); }
                    else if (dir == 1) { delta = Delta(0, -1); }
                    else if (dir == 2) { delta = Delta(-1, 0); }
                    else if (dir == 3) { delta = Delta(1, 0); }

                    Point newPoint = oldPoint + delta;
                    if (newPoint.m_x >= width ||
                        newPoint.m_y >= height ||
                        board[newPoint] != '#' ||
                        grid[newPoint].m_visited == v)
                    {
                        continue;
                    }

                    newPoints.push_back(newPoint);
                    grid[newPoint].m_visited = v;
                    grid[newPoint].m_point = oldPoint;
                    foundCount++;
                }
            }

            oldPoints.clear();
            oldPoints.swap(newPoints);
        }

        if (foundCount != outCount)
        {
            if (pMsg) *pMsg = "Obstacle surrounded";
            return false;
        }
    }

    {
        std::uniform_int_distribution<uint32_t> dist(0, (uint32_t)allIn.size() - 1);

        auto getEmpty = [&]() -> Point
        {
            while (true)
            {
                Point point = allIn[dist(gen)];
                if (board[point] == ' ') return point;
            }
        };

        for (uint32_t i = 0; i < puzzle.m_mNum; i++)
        {
            board[getEmpty()] = 'B';
        }
        for (uint32_t i = 0; i < puzzle.m_fNum; i++)
        {
            board[getEmpty()] = 'F';
        }
        for (uint32_t i = 0; i < puzzle.m_dNum; i++)
        {
            board[getEmpty()] = 'L';
        }
        for (uint32_t i = 0; i < puzzle.m_rNum; i++)
        {
            board[getEmpty()] = 'R';
        }
        for (uint32_t i = 0; i < puzzle.m_cNum; i++)
        {
            board[getEmpty()] = 'C';
        }
        for (uint32_t i = 0; i < puzzle.m_xNum; i++)
        {
            board[getEmpty()] = 'X';
        }
        *pStartPoint = getEmpty();
    }

    *pBoard = std::move(board);
    return true;
}
