#ifndef GRID_HPP
#define GRID_HPP

#include "Common.hpp"

#define FAST_GRID 1

template<typename T>
class Grid
{
public:
    Grid() :
        m_data(),
#if FAST_GRID
        m_shift(0),
#endif
        m_width(0),
        m_height(0)
    {
    }

    Grid(size_t width, size_t height) :
        m_data(),
#if FAST_GRID
        m_shift(0),
#endif
        m_width(0),
        m_height(0)
    {
        resize(width, height);
    }

    Grid(const Grid& other) :
        m_data(),
#if FAST_GRID
        m_shift(0),
#endif
        m_width(0),
        m_height(0)
    {
        *this = other;
    }

    Grid(Grid&& other) :
        m_data(std::move(other.m_data)),
#if FAST_GRID
        m_shift(other.m_shift),
#endif
        m_width(other.m_width),
        m_height(other.m_height)
    {
#if FAST_GRID
        other.m_shift = 0;
#endif
        other.m_width = 0;
        other.m_height = 0;
    }

    ~Grid() {}

    Grid& operator=(const Grid& other)
    {
        if (this != &other)
        {
            size_t width = other.m_width;
            size_t height = other.m_height;
            if (width == 0 || height == 0)
            {
                m_data.reset(nullptr);
#if FAST_GRID
                m_shift = 0;
#endif
                m_width = 0;
                m_height = 0;
            }
            else
            {
#if FAST_GRID
                size_t shift = other.m_shift;
                m_data.reset(new T[((size_t)1 << shift) * height]);
                const T* otherData = other.m_data.get();
                T* data = m_data.get();
                for (size_t y = 0; y < height; y++)
                {
                    for (size_t x = 0; x < width; x++)
                    {
                        data[(y << shift) + x] = otherData[(y << shift) + x];
                    }
                }
                m_shift = shift;
#else
                m_data.reset(new T[width * height]);
                size_t size = width * height;
                const T* otherData = other.m_data.get();
                T* data = m_data.get();
                for (size_t i = 0; i < size; i++)
                {
                    data[i] = otherData[i];
                }
#endif
                m_width = width;
                m_height = height;
            }
        }
        return *this;
    }
    Grid& operator=(Grid&& other)
    {
        if (this != &other)
        {
            m_data = std::move(other.m_data);
#if FAST_GRID
            m_shift = other.m_shift;
#endif
            m_width = other.m_width;
            m_height = other.m_height;
#if FAST_GRID
            other.m_shift = 0;
#endif
            other.m_width = 0;
            other.m_height = 0;
        }
        return *this;
    }

    size_t getWidth() const { return m_width; }
    size_t getHeight() const { return m_height; }

    void resize(size_t width, size_t height)
    {
        if (width == 0 || height == 0)
        {
            m_data.reset(nullptr);
#if FAST_GRID
            m_shift = 0;
#endif
            m_width = 0;
            m_height = 0;
        }
        else
        {
#if FAST_GRID
            size_t shift = 0;
            while (((size_t)1 << shift) < width)
            {
                shift++;
            }
            m_data.reset(new T[((size_t)1 << shift) * height]());
            m_shift = shift;
#else
            m_data.reset(new T[width * height]());
#endif
            m_width = width;
            m_height = height;
        }
    }

    void clear()
    {
        m_data.reset(nullptr);
#if FAST_GRID
        m_shift = 0;
#endif
        m_width = 0;
        m_height = 0;
    }

    const T& operator()(size_t x, size_t y) const
    {
#if FAST_GRID
        return m_data[(y << m_shift) + x];
#else
        return m_data[y * m_width + x];
#endif
    }

    T& operator()(size_t x, size_t y)
    {
#if FAST_GRID
        return m_data[(y << m_shift) + x];
#else
        return m_data[y * m_width + x];
#endif
    }

    template<typename Pos>
    const T& operator[](const Pos& pos) const
    { using std::get; return operator()(get<0>(pos), get<1>(pos)); }

    template<typename Pos>
    T& operator[](const Pos& pos)
    { using std::get; return operator()(get<0>(pos), get<1>(pos)); }

    void swap(Grid& other)
    {
        using std::swap;
        swap(m_data, other.m_data);
#if FAST_GRID
        swap(m_shift, other.m_shift);
#endif
        swap(m_width, other.m_width);
        swap(m_height, other.m_height);
    }

private:
    std::unique_ptr<T[]> m_data;
#if FAST_GRID
    size_t m_shift;
#endif
    size_t m_width;
    size_t m_height;
};

template<typename T>
void swap(Grid<T>& x, Grid<T>& y)
{
    x.swap(y);
}

#endif
