#include "Improver.hpp"
#include "Solver.hpp"
#include "TimeUtils.hpp"
#include <random>

using std::vector;
using std::pair;
using std::unique_ptr;

void improve(const Board& board,
             Point startPoint,
             const std::string& buy,
             uint32_t durationMS,
             uint32_t seed1,
             uint32_t seed2,
             bool verbose,
             Solution* pSolution,
             uint32_t* pTime)
{
    uint64_t startTimeMS = getTimeMS();

    uint32_t width = board.getWidth();
    uint32_t height = board.getHeight();

    std::default_random_engine gen;
    gen.seed(seed1);
    std::uniform_int_distribution<uint8_t> noiseDist(0, 255);
    std::uniform_int_distribution<int8_t> biasDist(-1, 1);

    uint32_t numStages = 5;
    uint32_t bonusStages = 15;
    uint32_t numPlans = 1 << (numStages - 1);
    typedef pair<unique_ptr<Plan>, uint32_t> PlanPair;
    vector<PlanPair> plans(numPlans);
    for (uint32_t iPlan = 0; iPlan < numPlans; iPlan++)
    {
        plans[iPlan].first.reset(new Plan);
        auto& plan = *plans[iPlan].first;

        plan.m_grid.resize(width, height);

        for (uint32_t y = 0; y < height; y++)
        {
            for (uint32_t x = 0; x < width; x++)
            {
                auto& cell = plan.m_grid(x, y);
                cell.m_noise = noiseDist(gen);
                cell.m_bias = 0;
            }
        }
    }

    gen.seed(seed2);

    uint32_t bestTime = 0;
    Solution bestSolution;

    for (uint32_t iPlan = 0; iPlan < numPlans; iPlan++)
    {
        auto& plan = *plans[iPlan].first;
        Solver solver(board, startPoint, buy, &plan);
        solver.run();
        uint32_t time = solver.m_time;
        plans[iPlan].second = time;
        bool better = iPlan == 0 || time < bestTime;
        bool same = iPlan != 0 && time == bestTime;
        if (better)
        {
            bestTime = time;
            bestSolution = std::move(solver.m_solution);
        }
        if (verbose)
        {
            printf("Init  %6" PRIu32 " %6" PRIu32 "%s\n",
                   time,
                   bestTime,
                   better ? "  NEW BEST!" : same ? "  (same)" : "");
        }
    }

    vector<Point> cells;
    for (uint32_t y = 0; y < height; y++)
    {
        for (uint32_t x = 0; x < width; x++)
        {
            if (board(x, y) != '#')
            {
                cells.emplace_back(x, y);
            }
        }
    }
    uint32_t numCells = cells.size();

    std::uniform_int_distribution<uint32_t> cellDist(0, numCells - 1);

    //uint32_t numChanges = numCells / 128;
    //if (numChanges < 1) numChanges = 1;
    uint32_t numNoiseChanges = 5;
    uint32_t numBiasChanges = 5;

    vector<pair<Point, uint8_t>> noiseChanges(numNoiseChanges);
    vector<pair<Point, int8_t>> biasChanges(numBiasChanges);

    uint64_t initTimeMS = getTimeMS();
    uint32_t remainingDurationMS = 0;
    if (initTimeMS - startTimeMS < durationMS)
    {
        remainingDurationMS = durationMS - (initTimeMS - startTimeMS);
    }

    vector<uint32_t> stageDurationsMS(numStages);
    vector<uint64_t> stageStartTimesMS(numStages);
    uint32_t tmpStageDurationMS = remainingDurationMS / (numStages + bonusStages);
    for (uint32_t iStage = 0; iStage < numStages; iStage++)
    {
        if (iStage == 0)
        {
            stageStartTimesMS[iStage] = initTimeMS;
        }
        else
        {
            stageStartTimesMS[iStage] = stageStartTimesMS[iStage - 1] + stageDurationsMS[iStage - 1];
        }

        if (iStage == numStages - 1)
        {
            stageDurationsMS[iStage] = tmpStageDurationMS * (1 + bonusStages);
        }
        else
        {
            stageDurationsMS[iStage] = tmpStageDurationMS;
        }
    }

    for (uint32_t iStage = 0; iStage < numStages; iStage++)
    {
        uint64_t stageStartTimeMS = stageStartTimesMS[iStage];
        uint32_t stageDurationMS = stageDurationsMS[iStage];
        uint32_t curNumPlans = numPlans >> iStage;

        while (getTimeMS() - stageStartTimeMS < stageDurationMS)
        {
            for (uint32_t iPlan = 0; iPlan < curNumPlans; iPlan++)
            {
                if (getTimeMS() - stageStartTimeMS >= stageDurationMS)
                {
                    break;
                }

                auto& plan = *plans[iPlan].first;

                for (uint32_t iChange = 0; iChange < numNoiseChanges; iChange++)
                {
                    Point point = cells[cellDist(gen)];
                    auto& change = noiseChanges[iChange];
                    change.first = point;
                    auto& cell = plan.m_grid[point];
                    change.second = cell.m_noise;
                    cell.m_noise = noiseDist(gen);
                }

                for (uint32_t iChange = 0; iChange < numBiasChanges; iChange++)
                {
                    Point point = cells[cellDist(gen)];
                    auto& change = biasChanges[iChange];
                    change.first = point;
                    auto& cell = plan.m_grid[point];
                    change.second = cell.m_bias;
                    cell.m_bias = biasDist(gen);
                }

                Solver solver(board, startPoint, buy, &plan);
                solver.run();
                uint32_t time = solver.m_time;
                bool planBetter = time < plans[iPlan].second;
                bool planSame = time == plans[iPlan].second;
                if (planBetter)
                {
                    plans[iPlan].second = time;
                }
                bool better = time < bestTime;
                bool same = time == bestTime;
                if (better)
                {
                    bestTime = time;
                    bestSolution = std::move(solver.m_solution);
                }
                if (verbose)
                {
                    printf("%4" PRIu32 "  %6" PRIu32 " %6" PRIu32 "%s\n",
                           curNumPlans,
                           time,
                           bestTime,
                           better ? "  NEW BEST!" : same ? "  (same)" : "");
                }
                if (planBetter || planSame)
                {
                    continue;
                }

                for (uint32_t iChange = numBiasChanges; iChange > 0; )
                {
                    iChange--;
                    auto& change = biasChanges[iChange];
                    auto& cell = plan.m_grid[change.first];
                    cell.m_bias = change.second;
                }

                for (uint32_t iChange = numNoiseChanges; iChange > 0; )
                {
                    iChange--;
                    auto& change = noiseChanges[iChange];
                    auto& cell = plan.m_grid[change.first];
                    cell.m_noise = change.second;
                }
            }
        }

        std::sort(plans.begin(), plans.begin() + curNumPlans,
                  [](const PlanPair& a, const PlanPair& b) -> bool
                  { return a.second < b.second; }
            );
    }

    *pSolution = std::move(bestSolution);
    *pTime = bestTime;
}
