#include "Problem.hpp"
#include "StringUtils.hpp"

using std::string;
using std::vector;

namespace
{
    bool parseUInt(const string& data,
                   size_t& pos,
                   uint32_t *pVal,
                   string* pMsg = nullptr)
    {
        size_t origPos = pos;
        size_t size = data.size();
        uint32_t val = 0;
        while (true)
        {
            if (pos >= size) break;
            char ch = data[pos];
            if (ch < '0' || ch > '9') break;
            if (val >= 100000000)
            {
                if (pMsg) *pMsg = "Number too large";
                return false;
            }
            val *= 10;
            val += (ch - '0');
            pos++;
        }
        if (pos == origPos)
        {
            if (pMsg) *pMsg = "No digits in number";
            return false;
        }
        *pVal = val;
        return true;
    }

    bool expectChar(const string& data,
                    size_t& pos,
                    char expectedCh,
                    string* pMsg = nullptr)
    {
        size_t size = data.size();
        if (pos >= size)
        {
            if (pMsg) *pMsg = string("Expected '") + expectedCh + "' at end of input";
            return false;
        }

        if (data[pos] != expectedCh)
        {
            if (pMsg) *pMsg = string("Expected '") + expectedCh + "'";
            return false;
        }

        pos++;
        return true;
    }

    bool parsePoint(const string& data,
                    size_t& pos,
                    Point* pPoint,
                    string* pMsg)
    {
        if (!expectChar(data, pos, '(', pMsg)) return false;

        uint32_t x = 0;
        if (!parseUInt(data, pos, &x, pMsg)) return false;

        if (!expectChar(data, pos, ',', pMsg)) return false;

        uint32_t y = 0;
        if (!parseUInt(data, pos, &y, pMsg)) return false;

        if (!expectChar(data, pos, ')', pMsg)) return false;

        *pPoint = Point(x, y);
        return true;
    }

    bool parsePoints(const string& data,
                     size_t& pos,
                     vector<Point>* pPoints,
                     string* pMsg)
    {
        pPoints->clear();
        size_t size = data.size();
        while (pos < size && data[pos] == '(')
        {
            Point point;
            if (!parsePoint(data, pos, &point, pMsg)) return false;
            pPoints->push_back(std::move(point));

            if (pos >= size || data[pos] != ',')
            {
                break;
            }
            pos++;
        }
        return true;
    }

    bool fillPolygon(Board& board,
                     const vector<Point>& polygon,
                     char val,
                     string* pMsg = nullptr)
    {
        vector<Point> edges;
        for (size_t i = 0; i < polygon.size(); i++)
        {
            auto& pointA = polygon[i];
            auto& pointB = polygon[(i + 1 == polygon.size()) ? 0 : i + 1];
            if (pointA.m_x != pointB.m_x) continue;
            uint32_t x = pointA.m_x;
            uint32_t minY = std::min(pointA.m_y, pointB.m_y);
            uint32_t maxY = std::max(pointA.m_y, pointB.m_y);
            for (uint32_t y = minY; y < maxY; y++)
            {
                edges.emplace_back(x, y);
            }
        }
        if (edges.size() % 2 != 0)
        {
            if (pMsg) *pMsg = "Odd number of vertical edges in polygon";
            return false;
        }
        std::sort(edges.begin(), edges.end(),
                  [](const Point& a, const Point& b) -> bool
                  { return a.m_y != b.m_y ? a.m_y < b.m_y : a.m_x < b.m_x; });
        for (size_t i = 0; i < edges.size(); i += 2)
        {
            auto& edgeA = edges[i];
            auto& edgeB = edges[i + 1];
            if (edgeA.m_y != edgeB.m_y)
            {
                if (pMsg) *pMsg = "Odd number of vertical edges in row";
                return false;
            }
            uint32_t y = edgeA.m_y;
            uint32_t minX = edgeA.m_x;
            uint32_t maxX = edgeB.m_x;
            for (uint32_t x = minX; x < maxX; x++)
            {
                board(x, y) = val;
            }
        }
        return true;
    }

    bool parseBooster(const string& data,
                      size_t& pos,
                      char* pCode,
                      Point* pPoint,
                      string* pMsg)
    {
        size_t size = data.size();
        if (pos >= size)
        {
            if (pMsg) *pMsg = "End of input while parsing booster code";
            return false;
        }
        char code = data[pos];
        if (code != 'B' &&
            code != 'F' &&
            code != 'L' &&
            code != 'X' &&
            code != 'R' &&
            code != 'C')
        {
            if (pMsg) *pMsg = string("Invalid booster code: '") + code + "'";
            return false;
        }
        pos++;
        *pCode = code;

        if (!parsePoint(data, pos, pPoint, pMsg)) return false;
        return true;
    }

    bool parseBoosters(const string& data,
                       size_t& pos,
                       vector<char>* pCodes,
                       vector<Point>* pPoints,
                       string* pMsg)
    {
        pCodes->clear();
        pPoints->clear();
        size_t size = data.size();
        while (pos < size &&
               data[pos] != '\r' &&
               data[pos] != '\n')
        {
            char code = 0;
            Point point;
            if (!parseBooster(data, pos, &code, &point, pMsg)) return false;
            pCodes->push_back(code);
            pPoints->push_back(std::move(point));

            if (pos >= size || data[pos] != ';')
            {
                break;
            }
            pos++;
        }
        return true;
    }
}

bool parseProblem(const string& data,
                  Board* pBoard,
                  Point* pStartPoint,
                  string* pMsg)
{
    size_t pos = 0;
    size_t size = data.size();

    vector<Point> points;
    if (!parsePoints(data, pos, &points, pMsg)) return false;

    if (points.empty())
    {
        if (pMsg) *pMsg = "Empty map";
        return false;
    }

    uint32_t width = 0;
    uint32_t height = 0;
    for (auto& point : points)
    {
        width = std::max(width, point.m_x);
        height = std::max(height, point.m_y);
    }

    auto& board = *pBoard;
    board.resize(width, height);

    for (uint32_t y = 0; y < height; y++)
    {
        for (uint32_t x = 0; x < width; x++)
        {
            board(x, y) = '#';
        }
    }

    if (!fillPolygon(board, points, ' ', pMsg)) return false;

    if (!expectChar(data, pos, '#', pMsg)) return false;

    if (!parsePoint(data, pos, pStartPoint, pMsg)) return false;

    if (!expectChar(data, pos, '#', pMsg)) return false;

    while (pos < size && data[pos] == '(')
    {
        if (!parsePoints(data, pos, &points, pMsg)) return false;
        if (!fillPolygon(board, points, '#', pMsg)) return false;

        if (pos >= size || data[pos] != ';')
        {
            break;
        }
        pos++;
    }

    if (!expectChar(data, pos, '#', pMsg)) return false;

    vector<char> codes;
    if (!parseBoosters(data, pos, &codes, &points, pMsg)) return false;

    if (codes.size() != points.size())
    {
        if (pMsg) *pMsg = "Internal error: booster codes and points size mismatch";
        return false;
    }

    size_t numBoosters = codes.size();
    for (size_t i = 0; i < numBoosters; i++)
    {
        char code = codes[i];
        uint32_t x = points[i].m_x;
        uint32_t y = points[i].m_y;
        if (x >= width || y >= height)
        {
            if (pMsg) *pMsg = "Booster outside of bounding box";
            return false;
        }
        if (board(x, y) != ' ')
        {
            if (pMsg)
            {
                *pMsg =
                    board(x, y) == '#' ?
                    "Booster outside of map or inside of obstacle" :
                    "Duplicate booster location";
            }
            return false;
        }
        board(x, y) = code;
    }

    while (pos < size &&
           (data[pos] == '\r' ||
            data[pos] == '\n'))
    {
        pos++;
    }

    if (pos < size)
    {
        if (pMsg) *pMsg = "Unexpected data at end of problem";
        return false;
    }

    return true;
}

namespace
{
    bool isVertex(const Board& board, Point point, Delta* pDelta, bool *pBad)
    {
        uint32_t width = board.getWidth();
        uint32_t height = board.getHeight();
        uint32_t cornerSW =
            (point.m_x - 1 < width &&
             point.m_y - 1 < height &&
             board(point.m_x - 1, point.m_y - 1) != '#');
        uint32_t cornerSE =
            (point.m_x < width &&
             point.m_y - 1 < height &&
             board(point.m_x, point.m_y - 1) != '#');
        uint32_t cornerNW =
            (point.m_x - 1 < width &&
             point.m_y < height &&
             board(point.m_x - 1, point.m_y) != '#');
        uint32_t cornerNE =
            (point.m_x < width &&
             point.m_y < height &&
             board(point.m_x, point.m_y) != '#');
        uint32_t cornerCount = cornerSW + cornerSE + cornerNW + cornerNE;
        if (cornerCount == 2 && cornerSW == cornerNE)
        {
            *pBad = true;
            return false;
        }
        if (cornerCount != 1 && cornerCount != 3)
        {
            *pBad = false;
            return false;
        }
        if (cornerNE && !cornerSE) *pDelta = Delta(1, 0);
        else if (cornerNW && !cornerNE) *pDelta = Delta(0, 1);
        else if (cornerSW && !cornerNW) *pDelta = Delta(-1, 0);
        else if (cornerSE && !cornerSW) *pDelta = Delta(0, -1);
        else
        {
            printf("isVertex: delta error\n");
            exit(1);
        }
        *pBad = false;
        return true;
    }

    string formatPoint(Point point)
    {
        return strprintf("(%" PRIu32 ",%" PRIu32 ")",
                         point.m_x,
                         point.m_y);
    }
}

bool formatProblem(const Board& board,
                   Point startPoint,
                   string* pData,
                   string* pMsg)
{
    uint32_t width = board.getWidth();
    uint32_t height = board.getHeight();
    bool first = true;
    Point firstVertex;
    Delta firstDelta;
    for (uint32_t y = 0; y <= height; y++)
    {
        for (uint32_t x = 0; x <= width; x++)
        {
            Delta delta;
            bool bBad = false;
            bool bVertex = isVertex(board, Point(x, y), &delta, &bBad);
            if (bBad)
            {
                if (pMsg) *pMsg = "Invalid vertex";
                return false;
            }
            if (bVertex)
            {
                if (first)
                {
                    first = false;
                    firstVertex = Point(x, y);
                    firstDelta = delta;
                }
            }
        }
    }

    string str;
    str += formatPoint(firstVertex);
    Point vertex = firstVertex;
    Delta delta = firstDelta;
    while (true)
    {
        vertex = vertex + delta;
        Delta curDelta;
        bool bCurBad = false;
        bool bCurVertex = isVertex(board, vertex, &curDelta, &bCurBad);
        if (bCurVertex)
        {
            if (vertex == firstVertex)
            {
                break;
            }
            str += ',';
            str += formatPoint(vertex);
            delta = curDelta;
        }
    }
    str += '#';
    str += formatPoint(startPoint);
    str += '#';
    // obstacles
    str += '#';
    bool firstBooster = true;
    for (uint32_t y = 0; y < height; y++)
    {
        for (uint32_t x = 0; x < width; x++)
        {
            char code = board(x, y);
            if (code != 'B' &&
                code != 'F' &&
                code != 'L' &&
                code != 'X' &&
                code != 'R' &&
                code != 'C')
            {
                continue;
            }
            if (firstBooster)
            {
                firstBooster = false;
            }
            else
            {
                str += ';';
            }
            str += code;
            str += formatPoint(Point(x, y));
        }
    }

    *pData = str;
    return true;
}
