#include "Process.hpp"
#include "TimeUtils.hpp"

#if PLATFORM_WINDOWS
#include <windows.h>
#else
#include <unistd.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <signal.h>
#endif

using std::string;
using std::vector;

class ProcessImpl
{
public:
    ProcessImpl() :
#if PLATFORM_WINDOWS
        m_hProcess(NULL),
        m_hIn(NULL),
        m_hOut(NULL)
#else
        m_pid(-1),
        m_fdIn(-1),
        m_fdOut(-1),
        m_blockingIn(true),
        m_blockingOut(true)
#endif
    {
    }

#if PLATFORM_WINDOWS
    HANDLE m_hProcess;
    HANDLE m_hIn;
    HANDLE m_hOut;
#else
    pid_t m_pid;
    int m_fdIn;
    int m_fdOut;
    bool m_blockingIn;
    bool m_blockingOut;
#endif
};

Process::Process() :
    m_pImpl(new ProcessImpl)
{
}

Process::Process(Process&& other) :
    m_pImpl(std::move(other.m_pImpl))
{
    other.m_pImpl.reset(new ProcessImpl);
}

Process::~Process()
{
    kill();
}

Process& Process::operator=(Process&& other)
{
    if (this != &other)
    {
        m_pImpl = std::move(other.m_pImpl);
        other.m_pImpl.reset(new ProcessImpl);
    }
    return *this;
}

void Process::swap(Process& other)
{
    using std::swap;
    swap(m_pImpl, other.m_pImpl);
}

#if PLATFORM_WINDOWS
namespace
{
    string escapeCommandLine(const string& program,
                             const vector<string>& args)
    {
        string result;
        for (int32_t iArg = -1; iArg < (int32_t)args.size(); iArg++)
        {
            const string& arg = (iArg == -1) ? program : args[iArg];
            if (iArg >= 0)
            {
                result.push_back(' ');
            }

            size_t len = arg.size();
            bool needQuotes = (len == 0);
            for (char ch : arg)
            {
                if (ch == ' ' ||
                    ch == '\t' ||
                    ch == '\n' ||
                    ch == '\v' ||
                    ch == '\"')
                {
                    needQuotes = true;
                    break;
                }
            }

            if (!needQuotes)
            {
                result.append(arg);
            }

            result.push_back('\"');
            size_t iCh = 0;
            while (iCh < len)
            {
                uint32_t backslashCount = 0;
                while (iCh < len && arg[iCh] == '\\')
                {
                    backslashCount++;
                    iCh++;
                }

                if (iCh == len)
                {
                    if (backslashCount > 0)
                    {
                        result.append(backslashCount * 2, '\\');
                    }
                }
                else
                {
                    char ch = arg[iCh];
                    iCh++;

                    if (ch == '\"')
                    {
                        result.append(backslashCount * 2 + 1, '\\');
                    }
                    else
                    {
                        if (backslashCount > 0)
                        {
                            result.append(backslashCount, '\\');
                        }
                    }
                    result.push_back(ch);
                }
            }
            result.push_back('\"');
        }

        return result;
    }
}
#endif

bool Process::create(const string& program,
                     const vector<string>& args,
                     bool swallowStdErr,
                     string* pMsg)
{
#if PLATFORM_WINDOWS
    HANDLE hIn[2] = {NULL, NULL};
    HANDLE hOut[2] = {NULL, NULL};

    SECURITY_ATTRIBUTES sa;
    sa.nLength = sizeof(sa);
    sa.lpSecurityDescriptor = NULL;
    sa.bInheritHandle = TRUE;

    if (!CreatePipe(&hIn[0], &hIn[1], &sa, 0))
    {
        if (pMsg)
        {
            *pMsg = "CreatePipe failed";
        }
        return false;
    }

    if (!SetHandleInformation(hIn[1], HANDLE_FLAG_INHERIT, 0))
    {
        CloseHandle(hIn[0]);
        CloseHandle(hIn[1]);
        if (pMsg)
        {
            *pMsg = "SetHandleInformation failed";
        }
        return false;
    }

    if (!CreatePipe(&hOut[0], &hOut[1], &sa, 0))
    {
        CloseHandle(hIn[0]);
        CloseHandle(hIn[1]);
        if (pMsg)
        {
            *pMsg = "CreatePipe failed";
        }
        return false;
    }

    if (!SetHandleInformation(hOut[0], HANDLE_FLAG_INHERIT, 0))
    {
        CloseHandle(hIn[0]);
        CloseHandle(hIn[1]);
        CloseHandle(hOut[0]);
        CloseHandle(hOut[1]);
        if (pMsg)
        {
            *pMsg = "SetHandleInformation failed";
        }
        return false;
    }

    string commandLine = escapeCommandLine(program, args);

    SIZE_T size = 0;
    if (InitializeProcThreadAttributeList(NULL, 1, 0, &size) ||
        GetLastError() != ERROR_INSUFFICIENT_BUFFER)
    {
        CloseHandle(hIn[0]);
        CloseHandle(hIn[1]);
        CloseHandle(hOut[0]);
        CloseHandle(hOut[1]);
        if (pMsg)
        {
            *pMsg = "InitializeProcThreadAttributeList failed";
        }
        return false;
    }

    LPPROC_THREAD_ATTRIBUTE_LIST pAL =
        (LPPROC_THREAD_ATTRIBUTE_LIST)HeapAlloc(GetProcessHeap(), 0, size);
    if (!pAL)
    {
        CloseHandle(hIn[0]);
        CloseHandle(hIn[1]);
        CloseHandle(hOut[0]);
        CloseHandle(hOut[1]);
        if (pMsg)
        {
            *pMsg = "HeapAlloc failed";
        }
        return false;
    }

    if (!InitializeProcThreadAttributeList(pAL, 1, 0, &size))
    {
        HeapFree(GetProcessHeap(), 0, pAL);
        CloseHandle(hIn[0]);
        CloseHandle(hIn[1]);
        CloseHandle(hOut[0]);
        CloseHandle(hOut[1]);
        if (pMsg)
        {
            *pMsg = "InitializeProcThreadAttributeList failed";
        }
        return false;
    }

    HANDLE inheritList[2] = { hIn[0], hOut[1] };
    if (!UpdateProcThreadAttribute(pAL,
                                   0,
                                   PROC_THREAD_ATTRIBUTE_HANDLE_LIST,
                                   inheritList,
                                   sizeof(inheritList),
                                   NULL,
                                   NULL))
    {
        DeleteProcThreadAttributeList(pAL);
        HeapFree(GetProcessHeap(), 0, pAL);
        CloseHandle(hIn[0]);
        CloseHandle(hIn[1]);
        CloseHandle(hOut[0]);
        CloseHandle(hOut[1]);
        if (pMsg)
        {
            *pMsg = "UpdateProcThreadAttribute failed";
        }
        return false;
    }

    STARTUPINFOEX siex;
    ZeroMemory(&siex, sizeof(STARTUPINFOEX));
    STARTUPINFO& si = siex.StartupInfo;
    si.cb = sizeof(STARTUPINFOEX);
    si.dwFlags = STARTF_USESTDHANDLES;
    si.hStdInput = hIn[0];
    si.hStdOutput = hOut[1];
    si.hStdError = NULL;
    siex.lpAttributeList = pAL;

    PROCESS_INFORMATION pi;
    ZeroMemory(&pi, sizeof(PROCESS_INFORMATION));

    if (!CreateProcess(NULL,
                       &commandLine[0],
                       NULL,
                       NULL,
                       TRUE,
                       EXTENDED_STARTUPINFO_PRESENT,
                       NULL,
                       NULL,
                       &si,
                       &pi))
    {
        DeleteProcThreadAttributeList(pAL);
        HeapFree(GetProcessHeap(), 0, pAL);
        CloseHandle(hIn[0]);
        CloseHandle(hIn[1]);
        CloseHandle(hOut[0]);
        CloseHandle(hOut[1]);
        if (pMsg)
        {
            *pMsg = "CreateProcess failed";
        }
        return false;
    }

    DeleteProcThreadAttributeList(pAL);
    HeapFree(GetProcessHeap(), 0, pAL);

    CloseHandle(pi.hThread);

    CloseHandle(hIn[0]);
    CloseHandle(hOut[1]);

    m_pImpl->m_hProcess = pi.hProcess;
    m_pImpl->m_hIn = hIn[1];
    m_pImpl->m_hOut = hOut[0];

    return true;
#else
    int fdsIn[2] = {0, 0};
    int fdsOut[2] = {0, 0};

    if (pipe2(fdsIn, O_CLOEXEC) != 0)
    {
        if (pMsg)
        {
            *pMsg = "pipe2 failed";
        }
        return false;
    }

    if (pipe2(fdsOut, O_CLOEXEC) != 0)
    {
        close(fdsIn[0]);
        close(fdsIn[1]);
        if (pMsg)
        {
            *pMsg = "pipe2 failed";
        }
        return false;
    }

    pid_t pid = fork();
    if (pid == -1)
    {
        close(fdsIn[0]);
        close(fdsIn[1]);
        close(fdsOut[0]);
        close(fdsOut[1]);
        if (pMsg)
        {
            *pMsg = "fork failed";
        }
        return false;
    }

    if (pid == 0)
    {
        // Child process

        if (setpgid(pid, 0) == -1)
        {
            fprintf(stderr, "setpgid failed\n");
            exit(1);
        }

        if (dup2(fdsIn[0], STDIN_FILENO) == -1 ||
            dup2(fdsOut[1], STDOUT_FILENO) == -1)
        {
            fprintf(stderr, "dup2 failed\n");
            exit(1);
        }

        close(fdsIn[0]);
        close(fdsIn[1]);
        close(fdsOut[0]);
        close(fdsOut[1]);

        if (swallowStdErr)
        {
            int fdErr = open("/dev/null", O_WRONLY);
            if (fdErr == -1)
            {
                fprintf(stderr, "open(\"/dev/null\") failed\n");
                exit(1);
            }

            if (dup2(fdErr, STDERR_FILENO) == -1)
            {
                fprintf(stderr, "dup2 failed\n");
                exit(1);
            }

            close(fdErr);
        }

        auto programCopy = program;
        auto argsCopy = args;
        vector<char*> argv(args.size() + 2);
        argv[0] = &programCopy[0];
        for (size_t iArg = 0; iArg < args.size(); iArg++)
        {
            argv[iArg + 1] = &argsCopy[iArg][0];
        }
        argv[args.size() + 1] = nullptr;

        execv(&programCopy[0], &argv[0]);

        fprintf(stderr, "execv failed\n");
        exit(1);
    }

    // Parent process

    if (setpgid(pid, 0) == -1)
    {
        // This is expected to fail after execv in the child
        //if (pMsg)
        //{
        //    *pMsg = "setpgid failed";
        //}
        //return false;
    }

    close(fdsIn[0]);
    close(fdsOut[1]);

    m_pImpl->m_pid = pid;
    m_pImpl->m_fdIn = fdsIn[1];
    m_pImpl->m_fdOut = fdsOut[0];
    m_pImpl->m_blockingIn = true;
    m_pImpl->m_blockingOut = true;

    return true;
#endif
}

bool Process::write(const void* buf,
                    size_t size,
                    size_t* pBytesWritten,
                    string* pMsg)
{
    return writeUntilMS(buf,
                        size,
                        pBytesWritten,
                        (uint64_t)-1,
                        nullptr,
                        pMsg);
}

bool Process::writeUntilMS(const void* buf,
                           size_t size,
                           size_t* pBytesWritten,
                           uint64_t timeMS,
                           bool* pTimeout,
                           string* pMsg)
{
#if PLATFORM_WINDOWS
    if (m_pImpl->m_hIn == NULL)
    {
        if (pMsg) *pMsg = "Write closed";
        return false;
    }

    if (size == 0)
    {
        if (pBytesWritten) *pBytesWritten = 0;
        if (pTimeout) *pTimeout = false;
        return true;
    }

    return false;
#else
    if (m_pImpl->m_fdIn == -1)
    {
        if (pMsg) *pMsg = "Write closed";
        return false;
    }

    if (size == 0)
    {
        if (pBytesWritten) *pBytesWritten = 0;
        if (pTimeout) *pTimeout = false;
        return true;
    }

    if (m_pImpl->m_blockingIn)
    {
        int flags = fcntl(m_pImpl->m_fdIn, F_GETFL, 0);
        if (flags == -1) flags = 0;
        if (fcntl(m_pImpl->m_fdIn, F_SETFL, flags | O_NONBLOCK) == -1)
        {
            if (pMsg) *pMsg = "fcntl failed";
            return false;
        }
        m_pImpl->m_blockingIn = false;
    }

    while (true)
    {
        ssize_t bytesWritten = ::write(m_pImpl->m_fdIn, buf, size);
        if (bytesWritten < 0)
        {
            int err = errno;
            if (err != EAGAIN && err != EWOULDBLOCK)
            {
                if (pMsg) *pMsg = "write failed";
                return false;
            }
        }
        else if ((size_t)bytesWritten > size)
        {
            if (pMsg) *pMsg = "write consumed too much data";
            return false;
        }
        else
        {
            if (pBytesWritten) *pBytesWritten = (size_t)bytesWritten;
            if (pTimeout) *pTimeout = false;
            return true;
        }

        if (timeMS != (uint64_t)-1 &&
            getTimeMS() >= timeMS)
        {
            if (pBytesWritten) *pBytesWritten = 0;
            if (pTimeout) *pTimeout = true;
            return true;
        }

        sleepMS(1);
    }
#endif
}

bool Process::writeForMS(const void* buf,
                         size_t size,
                         size_t* pBytesWritten,
                         uint64_t durationMS,
                         bool* pTimeout,
                         string* pMsg)
{
    return writeUntilMS(buf,
                        size,
                        pBytesWritten,
                        getTimeMS() + durationMS,
                        pTimeout,
                        pMsg);
}

void Process::writeClose()
{
#if PLATFORM_WINDOWS
    if (m_pImpl->m_hIn != NULL)
    {
        CloseHandle(m_pImpl->m_hIn);
        m_pImpl->m_hIn = NULL;
    }
#else
    if (m_pImpl->m_fdIn != -1)
    {
        close(m_pImpl->m_fdIn);
        m_pImpl->m_fdIn = -1;
    }
#endif
}

bool Process::read(void* buf,
                   size_t size,
                   size_t* pBytesRead,
                   string* pMsg)
{
    return readUntilMS(buf,
                       size,
                       pBytesRead,
                       (uint64_t)-1,
                       nullptr,
                       pMsg);
}

bool Process::readUntilMS(void* buf,
                          size_t size,
                          size_t* pBytesRead,
                          uint64_t timeMS,
                          bool* pTimeout,
                          string* pMsg)
{
#if PLATFORM_WINDOWS
    return false;
#else
    if (m_pImpl->m_fdOut == -1)
    {
        if (pMsg) *pMsg = "Read closed";
        return false;
    }

    if (m_pImpl->m_blockingOut)
    {
        int flags = fcntl(m_pImpl->m_fdOut, F_GETFL, 0);
        if (flags == -1) flags = 0;
        if (fcntl(m_pImpl->m_fdOut, F_SETFL, flags | O_NONBLOCK) == -1)
        {
            if (pMsg) *pMsg = "fcntl failed";
            return false;
        }
        m_pImpl->m_blockingOut = false;
    }

    while (true)
    {
        ssize_t bytesRead = ::read(m_pImpl->m_fdOut, buf, size);
        if (bytesRead < 0)
        {
            int err = errno;
            if (err != EAGAIN && err != EWOULDBLOCK)
            {
                if (pMsg) *pMsg = "read failed";
                return false;
            }
        }
        else if ((size_t)bytesRead > size)
        {
            if (pMsg) *pMsg = "read returned too much data";
            return false;
        }
        else
        {
            if (pBytesRead) *pBytesRead = (size_t)bytesRead;
            if (pTimeout) *pTimeout = false;
            return true;
        }

        if (timeMS != (uint64_t)-1 &&
            getTimeMS() >= timeMS)
        {
            if (pBytesRead) *pBytesRead = 0;
            if (pTimeout) *pTimeout = true;
            return true;
        }

        sleepMS(1);
    }
#endif
}

bool Process::readForMS(void* buf,
                        size_t size,
                        size_t* pBytesRead,
                        uint64_t durationMS,
                        bool* pTimeout,
                        string* pMsg)
{
    return readUntilMS(buf,
                       size,
                       pBytesRead,
                       getTimeMS() + durationMS,
                       pTimeout,
                       pMsg);
}

void Process::readClose()
{
#if PLATFORM_WINDOWS
    if (m_pImpl->m_hOut != NULL)
    {
        CloseHandle(m_pImpl->m_hOut);
        m_pImpl->m_hOut = NULL;
    }
#else
    if (m_pImpl->m_fdOut != -1)
    {
        close(m_pImpl->m_fdOut);
        m_pImpl->m_fdOut = -1;
    }
#endif
}

void Process::wait()
{
    waitUntilMS((uint64_t)-1,
                nullptr);
}

void Process::waitUntilMS(uint64_t timeMS,
                          bool* pTimeout)
{
#if PLATFORM_WINDOWS
    return;
#else
    if (m_pImpl->m_pid == -1)
    {
        if (pTimeout) *pTimeout = false;
        return;
    }

    while (true)
    {
        int status = 0;
        pid_t waitResult = waitpid(-m_pImpl->m_pid, &status, WNOHANG);
        if (waitResult == -1)
        {
            int err = errno;
            if (err == ECHILD)
            {
                writeClose();
                readClose();
                m_pImpl->m_pid = -1;
                if (pTimeout) *pTimeout = false;
                return;
            }

            fprintf(stderr, "waitpid failed\n");
            writeClose();
            readClose();
            m_pImpl->m_pid = -1;
            if (pTimeout) *pTimeout = false;
            return;
        }

        if (waitResult == 0)
        {
            if (timeMS != (uint64_t)-1 &&
                getTimeMS() >= timeMS)
            {
                if (pTimeout) *pTimeout = true;
                return;
            }

            sleepMS(1);
        }
    }
#endif
}

void Process::waitForMS(uint64_t durationMS,
                        bool* pTimeout)
{
    waitUntilMS(getTimeMS() + durationMS,
                pTimeout);
}

void Process::kill()
{
#if PLATFORM_WINDOWS
#else
    if (m_pImpl->m_pid == -1)
    {
        return;
    }

    if (::kill(-m_pImpl->m_pid, SIGTERM) == -1)
    {
        fprintf(stderr, "kill(SIGTERM) failed\n");
        writeClose();
        readClose();
        m_pImpl->m_pid = -1;
        return;
    }

    bool timeout = false;
    waitForMS(1000, &timeout);

    if (!timeout)
    {
        writeClose();
        readClose();
        m_pImpl->m_pid = -1;
        return;
    }

    fprintf(stderr, "Resorting to kill(SIGKILL)\n");

    if (::kill(-m_pImpl->m_pid, SIGKILL) == -1)
    {
        fprintf(stderr, "kill(SIGKILL) failed\n");
        writeClose();
        readClose();
        m_pImpl->m_pid = -1;
        return;
    }

    wait();
#endif
}

void swap(Process& x, Process& y)
{
    x.swap(y);
}
