#include "Sim.hpp"

using std::string;
using std::vector;

Sim::Sim() :
    m_board(),
    m_solution(),
    m_robots(),
    m_wrapped(),
    m_time(0),
    m_unwrapped(0),
    m_countB(0),
    m_countF(0),
    m_countL(0),
    m_countR(0),
    m_countC(0),
    m_invalid(false),
    m_done(false)
{
}

Sim::Sim(Board board,
         const Point& startPoint,
         const string& buy,
         Solution solution) :
    Sim()
{
    init(std::move(board), startPoint, buy, std::move(solution));
}

Sim::~Sim()
{
}

void Sim::init(Board board,
               const Point& startPoint,
               const string& buy,
               Solution solution)
{
    m_board = std::move(board);
    m_solution = std::move(solution);
    m_robots.clear();
    m_robots.resize(1);
    m_robots[0].m_point = startPoint;
    m_time = 0;
    m_invalid = false;
    m_done = false;

    uint32_t width = m_board.getWidth();
    uint32_t height = m_board.getHeight();
    for (uint32_t y = 0; y < height; y++)
    {
        for (uint32_t x = 0; x < width; x++)
        {
            if (m_board(x, y) != '#')
            {
                m_unwrapped++;
            }
        }
    }

    m_robots[0].m_deltas.emplace_back(0, 0);
    m_robots[0].m_deltas.emplace_back(1, 0);
    m_robots[0].m_deltas.emplace_back(1, 1);
    m_robots[0].m_deltas.emplace_back(1, -1);

    m_wrapped.resize(width, height);

    m_countB = 0;
    m_countF = 0;
    m_countL = 0;
    m_countR = 0;
    m_countC = 0;

    for (char code : buy)
    {
        if (code == 'B') m_countB++;
        else if (code == 'F') m_countF++;
        else if (code == 'L') m_countL++;
        else if (code == 'R') m_countR++;
        else if (code == 'C') m_countC++;
    }

    checkDone();
}

void Sim::step()
{
    if (m_invalid)
    {
        return;
    }

    m_time++;

    size_t numRobots = m_robots.size();
    for (size_t iRobot = 0; iRobot < numRobots; iRobot++)
    {
        auto& robot = m_robots[iRobot];

        bool activeF = false;
        if (robot.m_timeF != 0)
        {
            activeF = true;
            robot.m_timeF--;
        }

        bool activeL = false;
        if (robot.m_timeL != 0)
        {
            activeL = true;
            robot.m_timeL--;
        }

        if (robot.m_routeIdx == 0)
        {
            // Wrap on start or after clone
            wrap(robot);
        }
        pickUp(robot);

        Action action('Z');
        if (iRobot < m_solution.size() &&
            robot.m_routeIdx < (uint32_t)m_solution[iRobot].size())
        {
            action = m_solution[iRobot][robot.m_routeIdx];
        }
        robot.m_routeIdx++;

        char code = action.m_code;

        if (code == 'W' || code == 'S' || code == 'A' || code == 'D')
        {
            Delta delta;
            if (code == 'W') delta = Delta(0, 1);
            else if (code == 'S') delta = Delta(0, -1);
            else if (code == 'A') delta = Delta(-1, 0);
            else if (code == 'D') delta = Delta(1, 0);

            uint32_t numMoves = activeF ? 2 : 1;
            for (uint32_t iMove = 0; iMove < numMoves; iMove++)
            {
                Point newPoint = robot.m_point + delta;
                uint32_t width = m_board.getWidth();
                uint32_t height = m_board.getHeight();
                if (newPoint.m_x >= width ||
                    newPoint.m_y >= height)
                {
                    if (iMove == 0)
                    {
                        m_invalid = true;
                        return;
                    }
                    else
                    {
                        break;
                    }
                }
                if (m_board[newPoint] == '#')
                {
                    if (activeL)
                    {
                        m_board[newPoint] = ' ';
                        m_unwrapped++;
                        //wrap(robot); // Not according to visualizer
                    }
                    else if (iMove == 0)
                    {
                        m_invalid = true;
                        return;
                    }
                    else
                    {
                        break;
                    }
                }
                if (iMove == 1)
                {
                    pickUp(robot);
                }
                robot.m_point = newPoint;
                wrap(robot);
            }
        }
        else if (code == 'Z')
        {
        }
        else if (code == 'E' || code == 'Q')
        {
            if (code == 'E')
            {
                for (auto& delta : robot.m_deltas)
                {
                    delta = cw(delta);
                }
            }
            else // (code == 'Q')
            {
                for (auto& delta : robot.m_deltas)
                {
                    delta = ccw(delta);
                }
            }

            wrap(robot);
        }
        else if (code == 'B')
        {
            if (m_countB == 0)
            {
                m_invalid = true;
                return;
            }
            m_countB--;
            bool duplicate = false;
            bool adjacent = false;
            Delta newDelta = action.getDelta();
            for (auto& delta : robot.m_deltas)
            {
                if (newDelta == delta)
                {
                    duplicate = true;
                    break;
                }
                if (newDelta == delta + Delta(1, 0) ||
                    newDelta == delta + Delta(-1, 0) ||
                    newDelta == delta + Delta(0, 1) ||
                    newDelta == delta + Delta(0, -1))
                {
                    adjacent = true;
                }
            }
            if (duplicate || !adjacent)
            {
                m_invalid = true;
                return;
            }
            robot.m_deltas.push_back(newDelta);
            wrap(robot);
        }
        else if (code == 'F')
        {
            if (m_countF == 0)
            {
                m_invalid = true;
                return;
            }
            m_countF--;
            robot.m_timeF += 50;
        }
        else if (code == 'L')
        {
            if (m_countL == 0)
            {
                m_invalid = true;
                return;
            }
            m_countL--;
            robot.m_timeL += 30;
        }
        else if (code == 'R')
        {
            if (m_countR == 0)
            {
                m_invalid = true;
                return;
            }
            m_countR--;
            if (m_board[robot.m_point] == 'X' ||
                m_board[robot.m_point] == 'T')
            {
                m_invalid = true;
                return;
            }
            m_board[robot.m_point] = 'T';
        }
        else if (code == 'T')
        {
            Point newPoint = action.getPoint();
            uint32_t width = m_board.getWidth();
            uint32_t height = m_board.getHeight();
            if (newPoint.m_x >= width ||
                newPoint.m_y >= height ||
                m_board[newPoint] != 'T')
            {
                m_invalid = true;
                return;
            }
            robot.m_point = newPoint;
            wrap(robot);
        }
        else if (code == 'C')
        {
            if (m_countC == 0)
            {
                m_invalid = true;
                return;
            }
            m_countC--;
            if (m_board[robot.m_point] != 'X')
            {
                m_invalid = true;
                return;
            }
            m_robots.emplace_back(); // invalidates robot
            auto& oldRobot = m_robots[iRobot];
            auto& newRobot = m_robots.back();
            newRobot.m_point = oldRobot.m_point;
            newRobot.m_deltas.emplace_back(0, 0);
            newRobot.m_deltas.emplace_back(1, 0);
            newRobot.m_deltas.emplace_back(1, 1);
            newRobot.m_deltas.emplace_back(1, -1);
            //wrap(newRobot); // not according to visualizer
        }
        else
        {
            m_invalid = true;
            return;
        }
    }

    checkDone();
}

void Sim::run()
{
    while (!m_done)
    {
        step();

        if (m_invalid)
        {
            return;
        }
    }
}

void Sim::pickUp(SimRobot& robot)
{
    char code = m_board[robot.m_point];

    if (code == 'B')
    {
        m_board[robot.m_point] = ' ';
        m_countB++;
    }
    else if (code == 'F')
    {
        m_board[robot.m_point] = ' ';
        m_countF++;
    }
    else if (code == 'L')
    {
        m_board[robot.m_point] = ' ';
        m_countL++;
    }
    else if (code == 'R')
    {
        m_board[robot.m_point] = ' ';
        m_countR++;
    }
    else if (code == 'C')
    {
        m_board[robot.m_point] = ' ';
        m_countC++;
    }
}

void Sim::wrap(SimRobot& robot)
{
    robot.m_reachable.clear();
    uint32_t width = m_board.getWidth();
    uint32_t height = m_board.getHeight();
    for (Delta delta : robot.m_deltas)
    {
        Point a = robot.m_point;
        Point b = a + delta;
        if (b.m_x >= width ||
            b.m_y >= height ||
            m_board[b] == '#')
        {
            continue;
        }
        bool negX = delta.m_x < 0;
        bool negY = delta.m_y < 0;
        uint32_t incX = (uint32_t)(negY ? -delta.m_y : delta.m_y);
        uint32_t incY = (uint32_t)(negX ? -delta.m_x : delta.m_x);
        uint32_t fooX = incX;
        uint32_t fooY = incY;
        bool blocked = false;
        while (a != b)
        {
            uint32_t fooMin = std::min(fooX, fooY);
            fooX -= fooMin;
            fooY -= fooMin;
            if (fooX == 0)
            {
                a.m_x += negX ? (uint32_t)-1 : (uint32_t)1;
                fooX = incX * 2;
            }
            if (fooY == 0)
            {
                a.m_y += negY ? (uint32_t)-1 : (uint32_t)1;
                fooY = incY * 2;
            }
            if (m_board[a] == '#')
            {
                blocked = true;
                break;
            }
        }
        if (blocked)
        {
            continue;
        }
        robot.m_reachable.push_back(b);
        if (!m_wrapped[b])
        {
            m_wrapped[b] = 1;
            m_unwrapped--;
        }
    }
}

void Sim::checkDone()
{
    size_t numRobots = m_robots.size();
    for (size_t iRobot = 0; iRobot < numRobots; iRobot++)
    {
        if (iRobot < m_solution.size() &&
            m_robots[iRobot].m_routeIdx < (uint32_t)m_solution[iRobot].size())
        {
            return;
        }
    }
    m_done = true;
}
