#include "Solution.hpp"
#include "StringUtils.hpp"

using std::string;

string formatAction(const Action& action)
{
    if (action.m_code == 'B')
    {
        return strprintf("%c(%" PRIi32 ",%" PRIi32 ")",
                         action.m_code,
                         (int32_t)action.m_x,
                         (int32_t)action.m_y);
    }
    else if (action.m_code == 'T')
    {
        return strprintf("%c(%" PRIu32 ",%" PRIu32 ")",
                         action.m_code,
                         action.m_x,
                         action.m_y);
    }
    else
    {
        return string(1, action.m_code);
    }
}

string formatRoute(const Route& route)
{
    string str;
    for (auto& action : route)
    {
        str += formatAction(action);
    }
    return str;
}

string formatSolution(const Solution& solution)
{
    string str;
    for (size_t i = 0; i < solution.size(); i++)
    {
        if (i != 0)
        {
            str += '#';
        }
        str += formatRoute(solution[i]);
    }
    str += '\n';
    return str;
}

namespace
{
    bool parseUInt(const string& data,
                   size_t& pos,
                   uint32_t *pVal,
                   string* pMsg = nullptr)
    {
        size_t origPos = pos;
        size_t size = data.size();
        uint32_t val = 0;
        while (true)
        {
            if (pos >= size) break;
            char ch = data[pos];
            if (ch < '0' || ch > '9') break;
            if (val >= 100000000)
            {
                if (pMsg) *pMsg = "Number too large";
                return false;
            }
            val *= 10;
            val += (ch - '0');
            pos++;
        }
        if (pos == origPos)
        {
            if (pMsg) *pMsg = "No digits in number";
            return false;
        }
        *pVal = val;
        return true;
    }

    bool parseInt(const string& data,
                  size_t& pos,
                  int32_t* pVal,
                  string* pMsg)
    {
        size_t size = data.size();
        bool neg = false;
        if (pos < size && data[pos] == '-')
        {
            neg = true;
            pos++;
        }
        uint32_t val = 0;
        if (!parseUInt(data, pos, &val, pMsg)) return false;
        *pVal = neg ? -(int32_t)val : (int32_t)val;
        return true;
    }

    bool expectChar(const string& data,
                    size_t& pos,
                    char expectedCh,
                    string* pMsg = nullptr)
    {
        size_t size = data.size();
        if (pos >= size)
        {
            if (pMsg) *pMsg = string("Expected '") + expectedCh + "' at end of file";
            return false;
        }

        if (data[pos] != expectedCh)
        {
            if (pMsg) *pMsg = string("Expected '") + expectedCh + "'";
            return false;
        }

        pos++;
        return true;
    }

    bool parsePoint(const string& data,
                    size_t& pos,
                    Point* pPoint,
                    string* pMsg)
    {
        if (!expectChar(data, pos, '(', pMsg)) return false;

        uint32_t x = 0;
        if (!parseUInt(data, pos, &x, pMsg)) return false;

        if (!expectChar(data, pos, ',', pMsg)) return false;

        uint32_t y = 0;
        if (!parseUInt(data, pos, &y, pMsg)) return false;

        if (!expectChar(data, pos, ')', pMsg)) return false;

        *pPoint = Point(x, y);
        return true;
    }

    bool parseDelta(const string& data,
                    size_t& pos,
                    Delta* pDelta,
                    string* pMsg)
    {
        if (!expectChar(data, pos, '(', pMsg)) return false;

        int32_t x = 0;
        if (!parseInt(data, pos, &x, pMsg)) return false;

        if (!expectChar(data, pos, ',', pMsg)) return false;

        int32_t y = 0;
        if (!parseInt(data, pos, &y, pMsg)) return false;

        if (!expectChar(data, pos, ')', pMsg)) return false;

        *pDelta = Delta(x, y);
        return true;
    }

    bool parseAction(const string& data,
                     size_t& pos,
                     Action* pAction,
                     string* pMsg)
    {
        size_t size = data.size();
        if (pos >= size)
        {
            if (pMsg) *pMsg = "End of input while parsing action code";
            return false;
        }

        char code = data[pos];
        if (code != 'W' &&
            code != 'S' &&
            code != 'A' &&
            code != 'D' &&
            code != 'Z' &&
            code != 'E' &&
            code != 'Q' &&
            code != 'B' &&
            code != 'F' &&
            code != 'L' &&
            code != 'R' &&
            code != 'T' &&
            code != 'C')
        {
            if (pMsg) *pMsg = string("Invalid action code: '") + code + "'";
            return false;
        }
        pos++;

        if (code == 'B')
        {
            Delta delta;
            if (!parseDelta(data, pos, &delta, pMsg)) return false;
            *pAction = Action(code, delta);
        }
        else if (code == 'T')
        {
            Point point;
            if (!parsePoint(data, pos, &point, pMsg)) return false;
            *pAction = Action(code, point);
        }
        else
        {
            *pAction = Action(code);
        }
        return true;
    }

    bool parseRoute(const string& data,
                    size_t& pos,
                    Route* pRoute,
                    string* pMsg)
    {
        pRoute->clear();
        size_t size = data.size();
        while (pos < size &&
               data[pos] != '#' &&
               data[pos] != '\r' &&
               data[pos] != '\n')
        {
            Action action;
            if (!parseAction(data, pos, &action, pMsg)) return false;
            pRoute->push_back(std::move(action));
        }
        return true;
    }
}

bool parseSolution(const string& data,
                   Solution* pSolution,
                   string* pMsg)
{
    pSolution->clear();
    size_t pos = 0;
    size_t size = data.size();

    while (true)
    {
        Route route;
        if (!parseRoute(data, pos, &route, pMsg)) return false;
        pSolution->push_back(std::move(route));

        if (pos >= size || data[pos] != '#')
        {
            break;
        }
        pos++;
    }

    while (pos < size &&
           (data[pos] == '\r' ||
            data[pos] == '\n'))
    {
        pos++;
    }

    if (pos < size)
    {
        if (pMsg) *pMsg = "Unexpected data at end of solution";
        return false;
    }

    return true;
}
