#include "Solver.hpp"

using std::string;
using std::vector;

Solver::Solver() :
    m_board(),
    m_solution(),
    m_robots(),
    m_wrapped(),
    m_wrapPoints(),
    m_pointsC(),
    m_pointsX(),
    m_pointsB(),
    m_pPlan(nullptr),
    m_time(0),
    m_unwrapped(0),
    m_countB(0),
    m_countF(0),
    m_countL(0),
    m_countR(0),
    m_countC(0),
    m_needMissionUpdate(false)
{
}

Solver::Solver(Board board,
               const Point& startPoint,
               const string& buy,
               const Plan* pPlan) :
    Solver()
{
    init(std::move(board), startPoint, buy, pPlan);
}

Solver::~Solver()
{
}

void Solver::init(Board board,
                  const Point& startPoint,
                  const string& buy,
                  const Plan* pPlan)
{
    m_board = std::move(board);
    m_pPlan = pPlan;
    m_solution.clear();
    m_solution.resize(1);
    m_robots.clear();
    m_robots.reserve(16);
    m_robots.resize(1);
    uint32_t width = m_board.getWidth();
    uint32_t height = m_board.getHeight();
    m_robots[0].m_visited++;
    m_robots[0].m_grid.resize(width, height);
    m_robots[0].m_missionGrid.resize(width, height);
    m_robots[0].m_oldPoints.push_back(startPoint);
    m_robots[0].m_grid[startPoint].m_visited = m_robots[0].m_visited;
    m_robots[0].m_grid[startPoint].m_code = 0;
    m_time = 0;

    m_pointsC.clear();
    m_pointsX.clear();
    m_pointsB.clear();

    for (uint32_t y = 0; y < height; y++)
    {
        for (uint32_t x = 0; x < width; x++)
        {
            char code = m_board(x, y);
            if (code != '#')
            {
                m_unwrapped++;
                if (code == 'C') m_pointsC.emplace_back(x, y);
                else if (code == 'X') m_pointsX.emplace_back(x, y);
                else if (code == 'B') m_pointsB.emplace_back(x, y);
            }
        }
    }

    m_robots[0].m_deltas.emplace_back(0, 0);
    m_robots[0].m_deltas.emplace_back(1, 0);
    m_robots[0].m_deltas.emplace_back(1, 1);
    m_robots[0].m_deltas.emplace_back(1, -1);
    m_robots[0].m_cwDeltas.resize(m_robots[0].m_deltas.size());
    m_robots[0].m_ccwDeltas.resize(m_robots[0].m_deltas.size());
    for (size_t i = 0; i < m_robots[0].m_deltas.size(); i++)
    {
        m_robots[0].m_cwDeltas[i] = cw(m_robots[0].m_deltas[i]);
        m_robots[0].m_ccwDeltas[i] = ccw(m_robots[0].m_deltas[i]);
    }

    m_wrapped.resize(width, height);

    m_countB = 0;
    m_countF = 0;
    m_countL = 0;
    m_countR = 0;
    m_countC = 0;

    for (char code : buy)
    {
        if (code == 'B') m_countB++;
        else if (code == 'F') m_countF++;
        else if (code == 'L') m_countL++;
        else if (code == 'R') m_countR++;
        else if (code == 'C') m_countC++;
    }

    pickUpStart(m_robots[0], startPoint);

    m_needMissionUpdate = true;
    updateMissions(m_robots[0]);
}

void Solver::step()
{
    m_time++;

    uint32_t width = m_board.getWidth();
    uint32_t height = m_board.getHeight();

    size_t numRobots = m_robots.size();
    for (size_t iRobot = 0; iRobot < numRobots; iRobot++)
    {
        auto& robot = m_robots[iRobot];

        bool activeF = false;
        if (robot.m_timeF != 0)
        {
            activeF = true;
            robot.m_timeF--;
            if (robot.m_timeF == 0)
            {
                robot.m_expiredF = true;
            }
        }

        bool activeL = false;
        if (robot.m_timeL != 0)
        {
            activeL = true;
            robot.m_timeL--;
        }

        if (!activeF && robot.m_expiredF)
        {
            robot.m_expiredF = false;
            robot.m_oldPoints.clear();
            for (uint32_t y = 0; y < height; y++)
            {
                for (uint32_t x = 0; x < width; x++)
                {
                    if (robot.m_grid(x, y).m_visited == robot.m_visited)
                    {
                        robot.m_oldPoints.emplace_back(x, y);
                    }
                }
            }
        }

        if (robot.m_oldPoints.empty())
        {
            if (activeF)
            {
                m_solution[iRobot].push_back(Action('Z'));
                continue;
            }
            else
            {
                printf("oldPoints empty\n");
                exit(1);
            }
        }

        if (m_solution[iRobot].empty())
        {
            // Wrap on start or after clone
            wrap(robot, robot.m_deltas, robot.m_oldPoints[0]);
        }
        pickUpFinish(robot);

        bool found = false;
        int32_t foundPriority = 0;
        Point foundPoint;
        Action action('Z');
        bool foundHasTransitPoint = false;
        Point foundTransitPoint;

        // Clone

        if (action == Action('Z') &&
            m_countC != 0)
        {
            for (auto oldPoint : robot.m_oldPoints)
            {
                if (m_board[oldPoint] == 'X')
                {
                    int32_t priority = 0;
                    if (!found || priority > foundPriority)
                    {
                        found = true;
                        foundPriority = priority;
                        foundPoint = oldPoint;
                    }
                }
            }
            if (found)
            {
                m_countC--;
                action = Action('C');
                robot.m_newPoints.push_back(foundPoint);
                if (m_countC == 0)
                {
                    m_needMissionUpdate = true;
                }
            }
        }

        // Add manipulator

        if (action == Action('Z') &&
            m_countB != 0)
        {
            uint32_t minDeltas = 0;
            for (size_t jRobot = 0; jRobot < numRobots; jRobot++)
            {
                uint32_t numDeltas = m_robots[jRobot].m_deltas.size();
                if (jRobot == 0 || numDeltas < minDeltas)
                {
                    minDeltas = numDeltas;
                }
            }
            if (robot.m_deltas.size() == minDeltas)
            {
                m_countB--;
                Delta delta = robot.m_deltas.back() + (robot.m_deltas[3] - robot.m_deltas[1]);
                robot.m_deltas.push_back(delta);
                robot.m_cwDeltas.push_back(cw(delta));
                robot.m_ccwDeltas.push_back(ccw(delta));
                action = Action('B', delta);
                for (auto oldPoint : robot.m_oldPoints)
                {
                    m_wrapPoints.clear();
                    wrap(robot, robot.m_deltas, oldPoint, true);
                    uint32_t wrapCount = m_wrapPoints.size();
                    if (wrapCount != 0)
                    {
                        int32_t priority = 0;
                        priority += (int32_t)wrapCount * 0x10000;
                        if (m_pPlan)
                        {
                            for (auto point : m_wrapPoints)
                            {
                                auto& cell = m_pPlan->m_grid[point];
                                priority += (int32_t)cell.m_noise;
                                priority += (int32_t)cell.m_bias * 0x1000000;
                            }
                        }
                        if (!found || priority > foundPriority)
                        {
                            found = true;
                            foundPriority = priority;
                            foundPoint = oldPoint;
                        }
                    }
                }
                if (found)
                {
                    robot.m_newPoints.push_back(foundPoint);
                }
                else
                {
                    robot.m_newPoints = robot.m_oldPoints;
                }
            }
        }

        // Fast wheels

        if (action == Action('Z') &&
            m_countF != 0 &&
            robot.m_timeF == 0)
        {
            m_countF--;
            robot.m_timeF = 50;
            action = Action('F');
            robot.m_newPoints = robot.m_oldPoints;
        }

        // Move or turn

        if (action == Action('Z'))
        {
            for (auto oldPoint : robot.m_oldPoints)
            {
                // Move

                for (uint32_t dir = 0; dir < 4; dir++)
                {
                    char code = 'Z';
                    Delta delta;
                    if (dir == 0) { code = 'W'; delta = Delta(0, 1); }
                    else if (dir == 1) { code = 'S'; delta = Delta(0, -1); }
                    else if (dir == 2) { code = 'A'; delta = Delta(-1, 0); }
                    else if (dir == 3) { code = 'D'; delta = Delta(1, 0); }

                    bool pickedUp = false;
                    Point newPoint = oldPoint;
                    bool invalidMove = false;
                    bool hasTransitPoint = false;
                    Point transitPoint;
                    bool missionCompleted = false;

                    m_wrapPoints.clear();

                    uint32_t numMoves = activeF ? 2 : 1;
                    for (uint32_t iMove = 0; iMove < numMoves; iMove++)
                    {
                        Point tmpPoint = newPoint + delta;
                        if (tmpPoint.m_x >= width ||
                            tmpPoint.m_y >= height)
                        {
                            if (iMove == 0)
                            {
                                invalidMove = true;
                                break;
                            }
                            else
                            {
                                break;
                            }
                        }
                        if (m_board[tmpPoint] == '#')
                        {
                            if (activeL)
                            {
                                printf("Drill not supported\n");
                                exit(1);
                            }
                            else if (iMove == 0)
                            {
                                invalidMove = true;
                                break;
                            }
                            else
                            {
                                break;
                            }
                        }
                        if (iMove == 1)
                        {
                            hasTransitPoint = true;
                            transitPoint = newPoint;
                        }
                        newPoint = tmpPoint;
                        if (m_board[newPoint] == 'B' ||
                            m_board[newPoint] == 'F' ||
                            m_board[newPoint] == 'L' ||
                            m_board[newPoint] == 'R' ||
                            m_board[newPoint] == 'C')
                        {
                            pickedUp = true;
                        }
                        wrap(robot, robot.m_deltas, newPoint, true, iMove == 1);
                        if (robot.m_onMission &&
                            robot.m_missionGrid[newPoint].m_dist == 0)
                        {
                            missionCompleted = true;
                        }
                    }

                    if (invalidMove) continue;

                    if (robot.m_onMission &&
                        !missionCompleted)
                    {
                        if (robot.m_missionGrid[newPoint].m_dist >=
                            robot.m_missionGrid[oldPoint].m_dist)
                        {
                            continue;
                        }
                    }

                    auto& cell = robot.m_grid[newPoint];
                    if (cell.m_visited == robot.m_visited) continue;
                    cell.m_visited = robot.m_visited;
                    cell.m_point = oldPoint;
                    cell.m_code = code;

                    robot.m_newPoints.push_back(newPoint);

                    uint32_t wrapCount = m_wrapPoints.size();
                    if (missionCompleted || pickedUp || wrapCount != 0)
                    {
                        int32_t priority = 0;
                        if (missionCompleted)
                        {
                            priority += (int32_t)0x40000000;
                        }
                        if (pickedUp)
                        {
                            priority += (int32_t)0x10000000;
                        }
                        if (wrapCount != 0)
                        {
                            priority += (int32_t)wrapCount * 0x10000;
                            if (m_pPlan)
                            {
                                for (auto point : m_wrapPoints)
                                {
                                    auto& cell = m_pPlan->m_grid[point];
                                    priority += (int32_t)cell.m_noise;
                                    priority += (int32_t)cell.m_bias * 0x1000000;
                                }
                            }
                        }
                        if (!found || priority > foundPriority)
                        {
                            found = true;
                            foundPriority = priority;
                            foundPoint = newPoint;
                            action = Action('Z');
                            foundHasTransitPoint = hasTransitPoint;
                            foundTransitPoint = transitPoint;
                        }
                    }
                }

                // Turn

                for (uint32_t turnDir = 0; turnDir < 2; turnDir++)
                {
                    char code = (turnDir == 0) ? 'E' : 'Q';
                    auto& deltas = (turnDir == 0) ? robot.m_cwDeltas : robot.m_ccwDeltas;

                    m_wrapPoints.clear();
                    wrap(robot, deltas, oldPoint, true);

                    uint32_t wrapCount = m_wrapPoints.size();
                    if (wrapCount != 0)
                    {
                        int32_t priority = 0;
                        priority += (int32_t)wrapCount * 0x10000;
                        if (m_pPlan)
                        {
                            for (auto point : m_wrapPoints)
                            {
                                auto& cell = m_pPlan->m_grid[point];
                                priority += (int32_t)cell.m_noise;
                                priority += (int32_t)cell.m_bias * 0x1000000;
                            }
                        }
                        if (!found || priority > foundPriority)
                        {
                            found = true;
                            foundPriority = priority;
                            foundPoint = oldPoint;
                            action = Action(code);
                            foundHasTransitPoint = false;
                        }
                    }
                }
            }

            if (found)
            {
                if (action == Action('E'))
                {
                    size_t numDeltas = robot.m_deltas.size();
                    for (size_t i = 0; i < numDeltas; i++)
                    {
                        robot.m_deltas[i] = cw(robot.m_deltas[i]);
                        robot.m_cwDeltas[i] = cw(robot.m_cwDeltas[i]);
                        robot.m_ccwDeltas[i] = cw(robot.m_cwDeltas[i]);
                    }
                }
                else if (action == Action('Q'))
                {
                    size_t numDeltas = robot.m_deltas.size();
                    for (size_t i = 0; i < numDeltas; i++)
                    {
                        robot.m_deltas[i] = ccw(robot.m_deltas[i]);
                        robot.m_cwDeltas[i] = ccw(robot.m_cwDeltas[i]);
                        robot.m_ccwDeltas[i] = ccw(robot.m_cwDeltas[i]);
                    }
                }
            }
        }

        m_solution[iRobot].push_back(action);

        if (found)
        {
            Point tmpPoint = foundPoint;
            uint32_t routeIdx = (uint32_t)m_solution[iRobot].size() - 1;
            while (true)
            {
                auto& cell = robot.m_grid[tmpPoint];
                char code = cell.m_code;
                if (code == 0) break;
                while (m_solution[iRobot][routeIdx].m_code != 'Z')
                {
                    routeIdx--;
                }
                m_solution[iRobot][routeIdx] = Action(code);
                routeIdx--;
                tmpPoint = cell.m_point;
            }

            robot.m_newPoints.clear();
            robot.m_newPoints.push_back(foundPoint);
            robot.m_visited++;
            robot.m_grid[foundPoint].m_visited = robot.m_visited;
            robot.m_grid[foundPoint].m_code = 0;

            bool missionCompleted = false;

            if (foundHasTransitPoint)
            {
                pickUpStart(robot, foundTransitPoint);
                pickUpFinish(robot);
                wrap(robot, robot.m_deltas, foundTransitPoint);
                if (robot.m_onMission &&
                    robot.m_missionGrid[foundTransitPoint].m_dist == 0)
                {
                    missionCompleted = true;
                }
            }

            pickUpStart(robot, foundPoint);
            wrap(robot, robot.m_deltas, foundPoint);
            if (robot.m_onMission &&
                robot.m_missionGrid[foundPoint].m_dist == 0)
            {
                missionCompleted = true;
            }

            if (missionCompleted)
            {
                robot.m_onMission = false;
            }

            if (iRobot == 0)
            {
                updateMissions(robot);
            }
        }

        robot.m_oldPoints.clear();
        robot.m_oldPoints.swap(robot.m_newPoints);

        if (action == Action('C'))
        {
            m_robots.emplace_back(); // invalidates robot
            m_solution.emplace_back();
            //auto& oldRobot = m_robots[iRobot];
            auto& newRobot = m_robots.back();
            newRobot.m_visited++;
            newRobot.m_grid.resize(width, height);
            //newRobot.m_missionGrid.resize(width, height);
            newRobot.m_oldPoints.push_back(foundPoint);
            newRobot.m_grid[foundPoint].m_visited = newRobot.m_visited;
            newRobot.m_grid[foundPoint].m_code = 0;
            newRobot.m_deltas.emplace_back(0, 0);
            newRobot.m_deltas.emplace_back(1, 0);
            newRobot.m_deltas.emplace_back(1, 1);
            newRobot.m_deltas.emplace_back(1, -1);
            newRobot.m_cwDeltas.resize(newRobot.m_deltas.size());
            newRobot.m_ccwDeltas.resize(newRobot.m_deltas.size());
            for (size_t i = 0; i < newRobot.m_deltas.size(); i++)
            {
                newRobot.m_cwDeltas[i] = cw(newRobot.m_deltas[i]);
                newRobot.m_ccwDeltas[i] = ccw(newRobot.m_deltas[i]);
            }
            //wrap(newRobot, newRobot.m_deltas, foundPoint); // Not according to visualizer
        }
    }
}

void Solver::run()
{
    while (m_unwrapped != 0)
    {
        step();
    }
}

void Solver::pickUpStart(SolverRobot& robot, Point point)
{
    if (robot.m_code != 0)
    {
        printf("pickUpStart: robot.m_code != 0\n");
        exit(1);
    }

    char code = m_board[point];

    if (code == 'B' ||
        code == 'F' ||
        code == 'L' ||
        code == 'R' ||
        code == 'C')
    {
        m_board[point] = ' ';
        robot.m_code = code;
    }

    if (code == 'C')
    {
        m_pointsC.erase(std::remove(m_pointsC.begin(), m_pointsC.end(), point), m_pointsC.end());
        m_needMissionUpdate = true;
    }
    if (code == 'B')
    {
        m_pointsB.erase(std::remove(m_pointsB.begin(), m_pointsB.end(), point), m_pointsB.end());
        if (m_pointsC.empty() &&
            (m_countC == 0 || m_pointsX.empty()))
        {
            m_needMissionUpdate = true;
        }
    }
}

void Solver::pickUpFinish(SolverRobot& robot)
{
    char code = robot.m_code;
    robot.m_code = 0;

    if (code == 'C' && m_countC == 0)
    {
        m_needMissionUpdate = true;
    }

    if (code == 'B') m_countB++;
    else if (code == 'F') m_countF++;
    else if (code == 'L') m_countL++;
    else if (code == 'R') m_countR++;
    else if (code == 'C') m_countC++;
}

void Solver::wrap(SolverRobot& robot,
                  const vector<Delta>& deltas,
                  Point point,
                  bool test,
                  bool test2)
{
    size_t origWrapCount = m_wrapPoints.size();
    uint32_t width = m_board.getWidth();
    uint32_t height = m_board.getHeight();
    for (Delta delta : deltas)
    {
        Point a = point;
        Point b = a + delta;
        if (b.m_x >= width ||
            b.m_y >= height ||
            m_board[b] == '#' ||
            m_wrapped[b])
        {
            continue;
        }
        if (test2)
        {
            bool duplicate = false;
            for (size_t i = 0; i < origWrapCount; i++)
            {
                if (b == m_wrapPoints[i])
                {
                    duplicate = true;
                    break;
                }
            }
            if (duplicate)
            {
                continue;
            }
        }
        bool negX = delta.m_x < 0;
        bool negY = delta.m_y < 0;
        uint32_t incX = (uint32_t)(negY ? -delta.m_y : delta.m_y);
        uint32_t incY = (uint32_t)(negX ? -delta.m_x : delta.m_x);
        uint32_t fooX = incX;
        uint32_t fooY = incY;
        bool blocked = false;
        while (a != b)
        {
            uint32_t fooMin = std::min(fooX, fooY);
            fooX -= fooMin;
            fooY -= fooMin;
            if (fooX == 0)
            {
                a.m_x += negX ? (uint32_t)-1 : (uint32_t)1;
                fooX = incX * 2;
            }
            if (fooY == 0)
            {
                a.m_y += negY ? (uint32_t)-1 : (uint32_t)1;
                fooY = incY * 2;
            }
            if (m_board[a] == '#')
            {
                blocked = true;
                break;
            }
        }
        if (blocked)
        {
            continue;
        }
        if (test)
        {
            m_wrapPoints.push_back(b);
        }
        else //if (!m_wrapped[b])
        {
            m_wrapped[b] = 1;
            m_unwrapped--;
        }
    }
}

void Solver::updateMissions(SolverRobot& robot)
{
    if (!m_needMissionUpdate)
    {
        return;
    }
    m_needMissionUpdate = false;

    for (auto point : m_pointsC)
    {
        robot.m_missionOldPoints.push_back(point);
    }

    if (m_countC != 0)
    {
        for (auto point : m_pointsX)
        {
            robot.m_missionOldPoints.push_back(point);
        }
    }

    if (robot.m_missionOldPoints.empty())
    {
        for (auto point : m_pointsB)
        {
            robot.m_missionOldPoints.push_back(point);
        }
    }

    if (robot.m_missionOldPoints.empty())
    {
        return;
    }

    robot.m_onMission = true;
    robot.m_missionVisited++;
    uint32_t dist = 0;
    uint32_t width = m_board.getWidth();
    uint32_t height = m_board.getHeight();
    while (!robot.m_missionOldPoints.empty())
    {
        for (auto oldPoint : robot.m_missionOldPoints)
        {
            auto& cell = robot.m_missionGrid[oldPoint];
            if (cell.m_visited == robot.m_missionVisited)
            {
                continue;
            }
            cell.m_visited = robot.m_missionVisited;
            cell.m_dist = dist;

            for (uint32_t dir = 0; dir < 4; dir++)
            {
                Point newPoint;
                if (dir == 0) newPoint = oldPoint + Delta(0, 1);
                else if (dir == 1) newPoint = oldPoint + Delta(0, -1);
                else if (dir == 2) newPoint = oldPoint + Delta(-1, 0);
                else if (dir == 3) newPoint = oldPoint + Delta(1, 0);

                if (newPoint.m_x >= width ||
                    newPoint.m_y >= height ||
                    m_board[newPoint] == '#')
                {
                    continue;
                }

                robot.m_missionNewPoints.push_back(newPoint);
            }
        }

        robot.m_missionOldPoints.clear();
        robot.m_missionOldPoints.swap(robot.m_missionNewPoints);
        dist++;
    }
}
