#include "Common.hpp"
#include "Process.hpp"
#include "TimeUtils.hpp"
#include "StringUtils.hpp"
#include "FileUtils.hpp"
#include "Problem.hpp"
#include "Solution.hpp"
#include "Puzzle.hpp"
#include "Sim.hpp"
#include "Solver.hpp"
#include "Builder.hpp"
#include "Ansi.hpp"
#include "PrintBoard.hpp"
#include "Improver.hpp"

using std::string;
using std::vector;

bool lambdaCli(const vector<string>& args,
               string* pOutput,
               string* pMsg = nullptr)
{
    uint64_t timeoutTimeMS = getTimeMS() + 10000;

    Process process;

    if (!process.create("./lambda-cli.py", args, false, pMsg))
    {
        return false;
    }

    pOutput->clear();

    while (true)
    {
        char buf[1024];
        size_t bytesRead = 0;
        bool timeout = false;
        if (!process.readUntilMS(buf, sizeof(buf), &bytesRead, timeoutTimeMS, &timeout, pMsg))
        {
            return false;
        }

        if (timeout)
        {
            if (pMsg) *pMsg = "Timeout";
            return false;
        }

        if (bytesRead == 0)
        {
            break;
        }

        pOutput->insert(pOutput->size(), buf, bytesRead);
    }

    bool timeout = false;
    process.waitUntilMS(timeoutTimeMS, &timeout);

    if (timeout)
    {
        process.kill();
    }

    while (!pOutput->empty() &&
           (pOutput->back() == '\r' ||
            pOutput->back() == '\n'))
    {
        pOutput->resize(pOutput->size() - 1);
    }

    return true;
}

int main(int argc, char *argv[])
{
    string msg;
    string prevBlock;
    //Ansi ansi;

    while (true)
    {
        sleepMS(1000);

        string block;
        {
            if (!lambdaCli({"getmininginfo", "block"}, &block, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }
        }

        if (block == prevBlock)
        {
            printf("Block %s: skipping\n", block.c_str());
            printf("\n");
            sleepMS(60000);
            continue;
        }

        string puzzleData;
        {
            if (!lambdaCli({"getmininginfo", "puzzle"}, &puzzleData, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }
        }

        string problemData;
        {
            if (!lambdaCli({"getmininginfo", "task"}, &problemData, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }
        }

        string blockCheck;
        {
            if (!lambdaCli({"getmininginfo", "block"}, &blockCheck, &msg))
            {
                printf("%s\n", msg.c_str());
                return 1;
            }
        }

        if (blockCheck != block)
        {
            printf("Block %s != %s\n", block.c_str(), blockCheck.c_str());
            printf("\n");
            continue;
        }

        string problemFile = strprintf("mine-problems/block-%s.desc", block.c_str());
        string puzzleFile = strprintf("mine-puzzles/block-%s.puz", block.c_str());
        string solutionFile = strprintf("mine-solutions/block-%s.sol", block.c_str());
        string buildFile = strprintf("mine-builds/block-%s.desc", block.c_str());

        bool filesAlreadyExist = false;
        {
            size_t solutionFileSize = 0;
            size_t buildFileSize = 0;
            if (getFileSize(solutionFile, &solutionFileSize) ||
                getFileSize(buildFile, &buildFileSize))
            {
                filesAlreadyExist = true;
            }
        }

        printf("Block %s\n", block.c_str());

        if (filesAlreadyExist)
        {
            prevBlock = block;
            printf("Files already exist\n");
            printf("\n");
            continue;
        }

        if (!filesAlreadyExist)
        {
            if (!writeFile(problemFile, problemData))
            {
                printf("Error writing problem file\n");
                return 1;
            }

            if (!writeFile(puzzleFile, puzzleData))
            {
                printf("Error writing puzzle file\n");
                return 1;
            }
        }

        Board board;
        Point startPoint;
        if (!parseProblem(problemData, &board, &startPoint, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Puzzle puzzle;
        if (!parsePuzzle(puzzleData, &puzzle, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        Board puzzleBoard;
        Point puzzleStartPoint;
        if (!build(puzzle, &puzzleBoard, &puzzleStartPoint, &msg))
        {
            printf("%s\n", msg.c_str());
            printf("\n");
            continue;
        }

        string puzzleProblemData;
        if (!formatProblem(puzzleBoard, puzzleStartPoint, &puzzleProblemData, &msg))
        {
            printf("%s\n", msg.c_str());
            printf("\n");
            continue;
        }

        //printf("%s\n", puzzleProblemData.c_str());

        /*
        {
            vector<Point> robotPoints(1, puzzleStartPoint);
            printBoard(ansi, puzzleBoard, robotPoints);
        }
        */

        uint32_t time = 0;
        Solution solution;

        {
            Solver solver(board, startPoint, "");
            solver.run();
            printf("Solver time:   %" PRIu32 "\n", solver.m_time);

            improve(board, startPoint, "", 10 * 60 * 1000, 1, 2, false, &solution, &time);
            printf("Improver time: %" PRIu32 "\n", time);

            if (solver.m_time < time)
            {
                time = solver.m_time;
                solution = std::move(solver.m_solution);
            }
        }

        Sim sim(board, startPoint, "", solution);
        sim.run();

        if (sim.m_invalid)
        {
            printf("Solution invalid\n");
            return 1;
        }

        if (sim.m_unwrapped != 0)
        {
            printf("Solution incomplete\n");
            return 1;
        }

        if (time != sim.m_time)
        {
            printf("Time mismatch\n");
            return 1;
        }

        string solutionData = formatSolution(solution);

        if (filesAlreadyExist)
        {
            prevBlock = block;
            printf("Files already exist\n");
            printf("\n");
            continue;
        }

        if (!writeFile(solutionFile, solutionData))
        {
            printf("Error writing solution file\n");
            return 1;
        }

        if (!writeFile(buildFile, puzzleProblemData))
        {
            printf("Error writing build file\n");
            return 1;
        }

#if 1
        string submitOutput;
        if (!lambdaCli({"submit", block, solutionFile, buildFile},
                       &submitOutput,
                       &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }

        printf("Submit output: '%s'\n", submitOutput.c_str());
#endif

        prevBlock = block;

        printf("\n");
    }

    return 0;
}
