#include "Common.hpp"
#include "StringUtils.hpp"
#include "FileUtils.hpp"
#include "Problem.hpp"
#include "Solution.hpp"
#include "Sim.hpp"
#include "Solver.hpp"
#include "Improver.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage: solve [<start> [<end> [<seconds> [<seed1> [<seed2> [<buy>]]]]]]\n");
}

int main(int argc, char *argv[])
{
    uint32_t startProblem = 1;
    uint32_t endProblem = 300;
    uint32_t durationS = 12;
    uint32_t seed1 = 1;
    uint32_t seed2 = 1;
    string buy;
    if (argc < 1 || argc > 7)
    {
        usage();
        return 1;
    }
    if (argc >= 2)
    {
        char* endptr = nullptr;
        startProblem = (uint32_t)strtoul(argv[1], &endptr, 10);
        if (*argv[1] == '\0' ||
            *endptr != '\0' ||
            startProblem < 1 ||
            startProblem > 300)
        {
            usage();
            return 1;
        }
    }
    if (argc >= 3)
    {
        char* endptr = nullptr;
        endProblem = (uint32_t)strtoul(argv[2], &endptr, 10);
        if (*argv[2] == '\0' ||
            *endptr != '\0' ||
            endProblem < startProblem ||
            endProblem > 300)
        {
            usage();
            return 1;
        }
    }
    if (argc >= 4)
    {
        char* endptr = nullptr;
        durationS = (uint32_t)strtoul(argv[3], &endptr, 10);
        if (*argv[3] == '\0' ||
            *endptr != '\0' ||
            //durationS < 1 ||
            durationS > 1000000)
        {
            usage();
            return 1;
        }
    }
    if (argc >= 5)
    {
        char* endptr = nullptr;
        seed1 = (uint32_t)strtoul(argv[4], &endptr, 10);
        if (*argv[4] == '\0' ||
            *endptr != '\0')
        {
            usage();
            return 1;
        }
    }
    if (argc >= 6)
    {
        char* endptr = nullptr;
        seed2 = (uint32_t)strtoul(argv[5], &endptr, 10);
        if (*argv[5] == '\0' ||
            *endptr != '\0')
        {
            usage();
            return 1;
        }
    }
    if (argc >= 7)
    {
        buy = argv[6];
        std::sort(buy.begin(), buy.end());
        for (char code : buy)
        {
            if (code != 'B' &&
                code != 'F' &&
                code != 'L' &&
                code != 'R' &&
                code != 'C')
            {
                usage();
                return 1;
            }
        }
    }

    uint32_t durationMS = durationS * 1000;

    string msg;

    uint32_t totalTime = 0;
    uint32_t bestTotalTime = 0;

    for (uint32_t iProblem = startProblem; iProblem <= endProblem; iProblem++)
    {
        string problemFile = strprintf("../problems/prob-%03" PRIu32 ".desc", iProblem);
        string solutionFile = strprintf("../solutions%s/prob-%03" PRIu32 ".sol", buy.c_str(), iProblem);

        string problemData;
        if (!readFile(problemFile, &problemData))
        {
            printf("Error reading problem file\n");
            return 1;
        }

        Board board;
        Point startPoint;
        if (!parseProblem(problemData, &board, &startPoint, &msg))
        {
            printf("%03" PRIu32 ": %s\n", iProblem, msg.c_str());
            return 1;
        }

        uint32_t bestTime = 999999;
        {
            string solutionData;
            if (readFile(solutionFile, &solutionData))
            {
                Solution solution;
                if (!parseSolution(solutionData, &solution, &msg))
                {
                    printf("%03" PRIu32 ": %s\n", iProblem, msg.c_str());
                    return 1;
                }

                Sim sim(board, startPoint, buy, solution);

                sim.run();

                if (sim.m_invalid)
                {
                    printf("%03" PRIu32 ": Best solution invalid\n", iProblem);
                    return 1;
                }

                if (sim.m_unwrapped != 0)
                {
                    printf("%03" PRIu32 ": Best solution incomplete\n", iProblem);
                    return 1;
                }

                bestTime = sim.m_time;
            }
        }

        uint32_t time = 0;
        Solution solution;

        if (durationMS == 0)
        {
            Solver solver(board, startPoint, buy);
            solver.run();
            time = solver.m_time;
            solution = std::move(solver.m_solution);
        }
        else
        {
            bool verbose = false;
            improve(board, startPoint, buy, durationMS, seed1, seed2, verbose, &solution, &time);
        }

        printf("%03" PRIu32 ": %6" PRIu32 "  %6" PRIu32 "  ",
               iProblem,
               bestTime,
               time);
        fflush(stdout);

        Sim sim(board, startPoint, buy, solution);

        sim.run();

        if (sim.m_invalid)
        {
            printf("%s", formatSolution(solution).c_str());
            printf("Solution invalid\n");
            return 1;
        }

        if (sim.m_unwrapped != 0)
        {
            printf("Solution incomplete\n");
            return 1;
        }

        if (time != sim.m_time)
        {
            printf("Solver time mismatch\n");
            return 1;
        }

        printf("Solution ok");
        bool better = time < bestTime;
        bool same = time == bestTime;
        if (better)
        {
            printf("  NEW BEST!");
        }
        else if (same)
        {
            printf("  (same)");
        }
        printf("\n");

        totalTime += time;
        bestTotalTime += std::min(time, bestTime);

#if 1
        if (better)
        {
            string solutionData = formatSolution(solution);

            if (!writeFile(solutionFile, solutionData))
            {
                printf("Error writing solution file\n");
                return 1;
            }
        }
#endif
    }

    printf("Total time: %" PRIu32 "\n", totalTime);
    printf("Best time:  %" PRIu32 "\n", bestTotalTime);

    return 0;
}
